import uuid


class ResultsAPI:
    """
    While results = None, repr will be calling the api to get the results or the estimated time to completion. When it does, results will be stored here, and the API will no longer be called.
    """

    def __init__(self, api_key: str, job_id: str) -> None:
        self.api_key = api_key
        self.job_id = job_id
        self.status = None
        self.results = None

    def _get_results(self) -> dict:
        # TODO: API call to get status & results
        return ("Running", None)

    def __repr__(self) -> str:
        if self.results is None:
            self.status, self.results = self._get_results()
        return f"Results(api_key={self.api_key}, job_id={self.job_id}, status={self.status}, results={self.results})"


def JobRunnerAPI(api_key: str, job_dict: dict) -> ResultsAPI:
    api_key = api_key
    job_dict = job_dict
    # TODO: API call to send job data and return job_id
    job_id = str(uuid.uuid4())
    # return a ResultsAPI instance with the job_id
    return ResultsAPI(api_key, job_id)



# MODEL PRICES
model_prices = {
    "gpt-3.5-turbo": {
        "prompt_tokens": 0.001 / 1000,
        "completion_tokens": 0.002 / 1000,
    },
    "gpt-4-1106-preview": {
        "prompt_tokens": 0.01 / 1000,
        "completion_tokens": 0.03 / 1000,
    },
    "llama2-70b-chat": {
        "prompt_tokens": 0.03 / 1000,
        "completion_tokens": 0.03 / 1000,
    },
}



from __future__ import annotations
from functools import wraps
import io
import asyncio
import json
import time
import inspect
from typing import Coroutine
from abc import ABC, abstractmethod, ABCMeta
from rich.console import Console
from rich.table import Table


from edsl.trackers.TrackerAPI import TrackerAPI
from queue import Queue
from typing import Any, Callable, Type, List
from edsl.data import CRUDOperations, CRUD
from edsl.exceptions import LanguageModelResponseNotJSONError
from edsl.language_models.schemas import model_prices
from edsl.utilities.decorators import sync_wrapper, jupyter_nb_handler

from edsl.language_models.repair import repair
from typing import get_type_hints

from edsl.exceptions.language_models import LanguageModelAttributeTypeError
from edsl.enums import LanguageModelType, InferenceServiceType

from edsl.Base import RichPrintingMixin, PersistenceMixin


def handle_key_error(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
            assert True == False
        except KeyError as e:
            # Handle the KeyError exception
            return f"""KeyError occurred: {e}. This is most likely because the model you are using 
            returned a JSON object we were not expecting."""

    return wrapper


class RegisterLanguageModelsMeta(ABCMeta):
    "Metaclass to register output elements in a registry i.e., those that have a parent"
    _registry = {}  # Initialize the registry as a dictionary
    REQUIRED_CLASS_ATTRIBUTES = ["_model_", "_parameters_", "_inference_service_"]

    def __init__(cls, name, bases, dct):
        super(RegisterLanguageModelsMeta, cls).__init__(name, bases, dct)
        # if name != "LanguageModel":
        if (model_name := getattr(cls, "_model_", None)) is not None:
            RegisterLanguageModelsMeta.check_required_class_variables(
                cls, RegisterLanguageModelsMeta.REQUIRED_CLASS_ATTRIBUTES
            )

            ## Check that model name is valid
            if not LanguageModelType.is_value_valid(model_name):
                acceptable_values = [item.value for item in LanguageModelType]
                raise LanguageModelAttributeTypeError(
                    f"""A LanguageModel's model must be one of {LanguageModelType} values, which are
                    {acceptable_values}. You passed {model_name}."""
                )

            if not InferenceServiceType.is_value_valid(
                inference_service := getattr(cls, "_inference_service_", None)
            ):
                acceptable_values = [item.value for item in InferenceServiceType]
                raise LanguageModelAttributeTypeError(
                    f"""A LanguageModel's model must have an _inference_service_ value from 
                    {acceptable_values}. You passed {inference_service}."""
                )

            # LanguageModel children have to implement the async_execute_model_call method
            RegisterLanguageModelsMeta.verify_method(
                candidate_class=cls,
                method_name="async_execute_model_call",
                expected_return_type=dict[str, Any],
                required_parameters=[("user_prompt", str), ("system_prompt", str)],
                must_be_async=True,
            )
            # LanguageModel children have to implement the parse_response method
            RegisterLanguageModelsMeta.verify_method(
                candidate_class=cls,
                method_name="parse_response",
                expected_return_type=str,
                required_parameters=[("raw_response", dict[str, Any])],
                must_be_async=False,
            )
            # breakpoint()
            # RegisterLanguageModelsMeta._registry[name] = cls
            RegisterLanguageModelsMeta._registry[model_name] = cls

    @classmethod
    def get_registered_classes(cls):
        return cls._registry

    @staticmethod
    def check_required_class_variables(
        candidate_class: LanguageModel, required_attributes: List[str] = None
    ):
        """Checks if a class has the required attributes
        >>> class M:
        ...     _model_ = "m"
        ...     _parameters_ = {}
        >>> RegisterLanguageModelsMeta.check_required_class_variables(M, ["_model_", "_parameters_"])
        >>> class M2:
        ...     _model_ = "m"
        >>> RegisterLanguageModelsMeta.check_required_class_variables(M2, ["_model_", "_parameters_"])
        Traceback (most recent call last):
        ...
        Exception: Class M2 does not have required attribute _parameters_
        """
        required_attributes = required_attributes or []
        for attribute in required_attributes:
            if not hasattr(candidate_class, attribute):
                raise Exception(
                    f"Class {candidate_class.__name__} does not have required attribute {attribute}"
                )

    @staticmethod
    def verify_method(
        candidate_class: LanguageModel,
        method_name: str,
        expected_return_type: Any,
        required_parameters: List[tuple[str, Any]] = None,
        must_be_async: bool = False,
    ):
        RegisterLanguageModelsMeta._check_method_defined(candidate_class, method_name)

        required_parameters = required_parameters or []
        method = getattr(candidate_class, method_name)
        signature = inspect.signature(method)

        RegisterLanguageModelsMeta._check_return_type(method, expected_return_type)

        if must_be_async:
            RegisterLanguageModelsMeta._check_is_coroutine(method)

        # Check the parameters
        params = signature.parameters
        for param_name, param_type in required_parameters:
            RegisterLanguageModelsMeta._verify_parameter(
                params, param_name, param_type, method_name
            )

    @staticmethod
    def _check_method_defined(cls, method_name):
        """Checks if a method is defined in a class
        >>> class M:
        ...     def f(self): pass
        >>> RegisterLanguageModelsMeta._check_method_defined(M, "f")
        >>> RegisterLanguageModelsMeta._check_method_defined(M, "g")
        Traceback (most recent call last):
        ...
        NotImplementedError: g method must be implemented
        """
        if not hasattr(cls, method_name):
            raise NotImplementedError(f"{method_name} method must be implemented")

    @staticmethod
    def _check_is_coroutine(func: Callable):
        """
        Checks to make sure it's a coroutine function
        >>> def f(): pass
        >>> RegisterLanguageModelsMeta._check_is_coroutine(f)
        Traceback (most recent call last):
        ...
        TypeError: A LangugeModel class with method f must be an asynchronous method
        """
        if not inspect.iscoroutinefunction(func):
            raise TypeError(
                f"A LangugeModel class with method {func.__name__} must be an asynchronous method"
            )

    @staticmethod
    def _verify_parameter(params, param_name, param_type, method_name):
        if param_name not in params:
            raise TypeError(
                f"""Parameter "{param_name}" of method "{method_name}" must be defined.
                """
            )
        if params[param_name].annotation != param_type:
            raise TypeError(
                f"""Parameter "{param_name}" of method "{method_name}" must be of type {param_type.__name__}.
                Got {params[param_name].annotation} instead.
                """
            )

    @staticmethod
    def _check_return_type(method, expected_return_type):
        """
        Checks if the return type of a method is as expected
        >>> class M:
        ...     async def f(self) -> str: pass
        >>> RegisterLanguageModelsMeta._check_return_type(M.f, str)
        >>> class N:
        ...     async def f(self) -> int: pass
        >>> RegisterLanguageModelsMeta._check_return_type(N.f, str)
        Traceback (most recent call last):
        ...
        TypeError: Return type of f must be <class 'str'>. Got <class 'int'>
        """
        if inspect.isroutine(method):
            # return_type = inspect.signature(method).return_annotation
            return_type = get_type_hints(method)["return"]
            if return_type != expected_return_type:
                raise TypeError(
                    f"Return type of {method.__name__} must be {expected_return_type}. Got {return_type}."
                )

    @classmethod
    def model_names_to_classes(cls):
        d = {}
        for classname, cls in cls._registry.items():
            if hasattr(cls, "_model_"):
                d[cls._model_] = cls
            else:
                raise Exception(
                    f"Class {classname} does not have a _model_ class attribute"
                )
        return d


class LanguageModel(
    RichPrintingMixin, PersistenceMixin, ABC, metaclass=RegisterLanguageModelsMeta
):
    """ABC for LLM subclasses."""

    _model_ = None

    def __init__(self, crud: CRUDOperations = CRUD, **kwargs):
        """
        Attributes:
        - all attributes inherited from subclasses
        - lock: lock for this model to ensure TODO
        - api_queue: queue that records messages about API calls the model makes. Used by `InterviewManager` to update details about state of model.
        """
        self.model = getattr(self, "_model_", None)
        default_parameters = getattr(self, "_parameters_", None)
        parameters = self._overide_default_parameters(kwargs, default_parameters)
        self.parameters = parameters

        for key, value in parameters.items():
            setattr(self, key, value)

        for key, value in kwargs.items():
            if key not in parameters:
                setattr(self, key, value)

        # for key, value in kwargs.items():
        # setattr(self, key, value)
        self.api_queue = Queue()
        self.crud = crud

    @staticmethod
    def _overide_default_parameters(passed_parameter_dict, default_parameter_dict):
        """Returns a dictionary of parameters, with passed parameters taking precedence over defaults.

        >>> LanguageModel._overide_default_parameters(passed_parameter_dict={"temperature": 0.5}, default_parameter_dict={"temperature":0.9})
        {'temperature': 0.5}
        >>> LanguageModel._overide_default_parameters(passed_parameter_dict={"temperature": 0.5}, default_parameter_dict={"temperature":0.9, "max_tokens": 1000})
        {'temperature': 0.5, 'max_tokens': 1000}
        """
        parameters = dict({})
        for parameter, default_value in default_parameter_dict.items():
            if parameter in passed_parameter_dict:
                parameters[parameter] = passed_parameter_dict[parameter]
            else:
                parameters[parameter] = default_value
        return parameters

    @abstractmethod
    async def async_execute_model_call():
        pass

    @jupyter_nb_handler
    def execute_model_call(self, *args, **kwargs) -> Coroutine:
        async def main():
            results = await asyncio.gather(
                self.async_execute_model_call(*args, **kwargs)
            )
            return results[0]  # Since there's only one task, return its result

        return main()

    @abstractmethod
    def parse_response(raw_response: dict[str, Any]) -> str:
        """Parses the API response and returns the response text.
        What is returned by the API is model-specific and often includes meta-data that we do not need.
        For example, here is the results from a call to GPT-4:

        {
            "id": "chatcmpl-8eORaeuVb4po9WQRjKEFY6w7v6cTm",
            "choices": [
                {
                    "finish_reason": "stop",
                    "index": 0,
                    "logprobs": None,
                    "message": {
                        "content": "Hello! How can I assist you today? If you have any questions or need information on a particular topic, feel free to ask.",
                        "role": "assistant",
                        "function_call": None,
                        "tool_calls": None,
                    },
                }
            ],
            "created": 1704637774,
            "model": "gpt-4-1106-preview",
            "object": "chat.completion",
            "system_fingerprint": "fp_168383a679",
            "usage": {"completion_tokens": 27, "prompt_tokens": 13, "total_tokens": 40},
        }

        To actually tract the response, we need to grab
            data["choices[0]"]["message"]["content"].
        """
        raise NotImplementedError

    def _update_response_with_tracking(
        self, response, start_time, cached_response=False
    ):
        end_time = time.time()
        response["elapsed_time"] = end_time - start_time
        response["timestamp"] = end_time
        self._post_tracker_event(response)
        response["cached_response"] = cached_response
        return response

    async def async_get_raw_response(
        self, user_prompt: str, system_prompt: str = ""
    ) -> dict[str, Any]:
        """This is some middle-ware that handles the caching of responses.
        If the cache isn't being used, it just returns a 'fresh' call to the LLM,
        but appends some tracking information to the response (using the _update_response_with_tracking method).
        But if cache is being used, it first checks the database to see if the response is already there.
        If it is, it returns the cached response, but again appends some tracking information.
        If it isn't, it calls the LLM, saves the response to the database, and returns the response with tracking information.

        If self.use_cache is True, then attempts to retrieve the response from the database;
        if not in the DB, calls the LLM and writes the response to the DB."""
        start_time = time.time()

        if not self.use_cache:
            response = await self.async_execute_model_call(user_prompt, system_prompt)
            return self._update_response_with_tracking(response, start_time, False)

        cached_response = self.crud.get_LLMOutputData(
            model=str(self.model),
            parameters=str(self.parameters),
            system_prompt=system_prompt,
            prompt=user_prompt,
        )

        if cached_response:
            response = json.loads(cached_response)
            cache_used = True
        else:
            response = await self.async_execute_model_call(user_prompt, system_prompt)
            self._save_response_to_db(user_prompt, system_prompt, response)
            cache_used = False

        return self._update_response_with_tracking(response, start_time, cache_used)

    get_raw_response = sync_wrapper(async_get_raw_response)

    def _save_response_to_db(self, prompt, system_prompt, response):
        try:
            output = json.dumps(response)
        except json.JSONDecodeError:
            raise LanguageModelResponseNotJSONError
        self.crud.write_LLMOutputData(
            model=str(self.model),
            parameters=str(self.parameters),
            system_prompt=system_prompt,
            prompt=prompt,
            output=output,
        )

    async def async_get_response(self, user_prompt: str, system_prompt: str = ""):
        """Get response, parse, and return as string."""
        raw_response = await self.async_get_raw_response(user_prompt, system_prompt)
        response = self.parse_response(raw_response)
        try:
            dict_response = json.loads(response)
        except json.JSONDecodeError as e:
            print("Could not load JSON. Trying to repair.")
            print(response)
            dict_response, success = await repair(response, str(e))
            if not success:
                raise Exception("Even the repair failed.")
        return dict_response

    get_response = sync_wrapper(async_get_response)

    #######################
    # USEFUL METHODS
    #######################
    def _post_tracker_event(self, raw_response: dict[str, Any]) -> None:
        """Parses the API response and sends usage details to the API Queue."""
        usage = raw_response.get("usage", {})
        usage.update(
            {
                "cached_response": raw_response.get("cached_response", None),
                "elapsed_time": raw_response.get("elapsed_time", None),
                "timestamp": raw_response.get("timestamp", None),
            }
        )
        event = TrackerAPI.APICallDetails(details=usage)
        self.api_queue.put(event)

    def cost(self, raw_response: dict[str, Any]) -> float:
        """Returns the dollar cost of a raw response."""
        keys = raw_response["usage"].keys()
        prices = model_prices.get(self.model)
        return sum([prices.get(key, 0.0) * raw_response["usage"][key] for key in keys])

    #######################
    # SERIALIZATION METHODS
    #######################
    def to_dict(self) -> dict[str, Any]:
        """Converts instance to a dictionary."""
        return {"model": self.model, "parameters": self.parameters}

    @classmethod
    def from_dict(cls, data: dict) -> Type[LanguageModel]:
        """Converts dictionary to a LanguageModel child instance."""
        from edsl.language_models.registry import get_model_class

        model_class = get_model_class(data["model"])
        data["use_cache"] = True
        return model_class(**data)

    #######################
    # DUNDER METHODS
    #######################
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(model = '{self.model}', parameters={self.parameters})"

    def __add__(self, other_model: Type[LanguageModel]) -> Type[LanguageModel]:
        """Combine two models into a single model (other_model takes precedence over self)"""
        print(
            f"""Warning: one model is replacing another. If you want to run both models, use a single `by` e.g., 
              by(m1, m2, m3) not by(m1).by(m2).by(m3)."""
        )
        return other_model or self

    def rich_print(self):
        """Displays an object as a table."""
        table = Table(title="Language Model")
        table.add_column("Attribute", style="bold")
        table.add_column("Value")

        to_display = self.__dict__.copy()
        for attr_name, attr_value in to_display.items():
            table.add_row(attr_name, repr(attr_value))

        return table

    @classmethod
    def example(cls):
        "Returns a default instance of the class"
        from edsl import Model

        return Model(Model.available()[0])


if __name__ == "__main__":
    # import doctest
    # doctest.testmod()
    from edsl.language_models import LanguageModel

    print(LanguageModel.example())



import os
import shutil
from contextlib import contextmanager
from sqlalchemy import create_engine, event
from sqlalchemy.exc import SQLAlchemyError, DatabaseError
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.sql import text
from edsl.config import Config, CONFIG
from edsl.data import Base
from edsl.exceptions import DatabaseConnectionError, DatabaseIntegrityError


class Database:
    """The Database class manages the connection to the database."""

    def __init__(self, config: Config = CONFIG):
        """Initializes the database connection."""
        self.database_path = config.get("EDSL_DATABASE_PATH")
        try:
            self.engine = create_engine(self.database_path)

            # listener sets WAL mode on connect
            # @event.listens_for(self.engine, "connect")
            # def set_wal_mode(dbapi_connection, connection_record):
            #     cursor = dbapi_connection.cursor()
            #     cursor.execute("PRAGMA journal_mode=WAL;")
            #     cursor.close()

            # test connection
            with self.engine.connect() as _:
                pass
            Base.metadata.create_all(bind=self.engine)
        except DatabaseError as e:
            if "malformed" in str(e):
                file_path, temp_path = self.paths
                if temp_path and os.path.exists(temp_path):
                    raise DatabaseIntegrityError(
                        f"Your database is malformed. "
                        f"Try replacing your main database file with the temp copy.\n"
                        f"Database file: {file_path}\n"
                        f"Temp file: {temp_path}"
                    )
                else:
                    raise DatabaseIntegrityError(
                        f"Your database is malformed. "
                        f"Please delete your database file.\n"
                        f"Database file: {file_path}\n"
                    )

        except SQLAlchemyError as e:
            raise DatabaseConnectionError(str(e)) from None
        self.SessionLocal = scoped_session(
            sessionmaker(autocommit=False, autoflush=False, bind=self.engine)
        )

    @contextmanager
    def get_db(self):
        """Generator that yields a database session."""
        db = None
        try:
            db = self.SessionLocal()
            yield db
        except SQLAlchemyError as e:
            raise DatabaseConnectionError(str(e)) from None
        finally:
            if db is not None:
                self.SessionLocal.remove()

    def _is_healthy(self):
        """Returns True if the database is healthy, False otherwise."""
        try:
            with self.engine.connect() as connection:
                result = connection.execute(text("PRAGMA integrity_check"))
                check_result = result.fetchone()
                if check_result[0] == "ok":
                    return True
                else:
                    return False
        except SQLAlchemyError as e:
            return False

    @property
    def paths(self) -> tuple:
        """Returns the (db_path, temp_path), or (None, None) if not applicable."""
        if (
            self.database_path.startswith("sqlite:///")
            and "memory" not in self.database_path
        ):
            file_path = self.database_path.replace("sqlite:///", "")
            temp_path = f"{file_path}_temp"
            return file_path, temp_path
        else:
            return None, None

    def _create_copy(self) -> None:
        """Creates a copy of the database."""
        file_path, temp_path = self.paths
        if file_path:
            try:
                shutil.copyfile(file_path, temp_path)
            except IOError as e:
                print(f"Error creating a copy of the database: {e}")
                raise e

    def _delete_copy(self) -> None:
        """Deletes a copy of the database."""
        file_path, temp_path = self.paths
        if file_path and os.path.exists(temp_path):
            os.remove(temp_path)

    def _health_check_pre_run(self):
        """Performs a health check before running a job."""
        # create a copy of the database if it is healthy
        if self._is_healthy():
            self._create_copy()
        else:
            file_path, temp_path = self.paths
            raise DatabaseIntegrityError(
                f"Could not run job because your database is not healthy.\n"
                f"Try replacing your main database file with the temp copy.\n"
                f"Database file: {file_path}\n"
                f"Temp file: {temp_path}"
            )

    def _health_check_post_run(self):
        """Performs a health check before running a job."""
        if self._is_healthy():
            self._delete_copy()
        else:
            file_path, temp_path = self.paths
            print(
                f"Running this job malformed your database.\n"
                f"Try replacing your main database file with the temp copy.\n"
                f"Database file: {file_path}\n"
                f"Temp file: {temp_path}"
            )


# Note: Python modules are singletons. As such, once `database` is imported
# the same instance of it is reused across the application.
database = Database()



from __future__ import annotations
import random
import textwrap
from jinja2 import Template
from typing import Any, Optional, Union
from edsl.questions import Question
from edsl.questions.descriptors import (
    IntegerDescriptor,
    QuestionOptionsDescriptor,
)
from edsl.scenarios import Scenario
from edsl.utilities import random_string


class QuestionCheckBox(Question):
    """
    This question asks the user to select options from a list.

    Arguments:
    - `question_name` is the name of the question (string)
    - `question_options` are the options the user should select from (list of strings)
    - `question_text` is the text of the question (string)

    Optional arguments:
    - `min_selections` is the minimum number of options that must be selected (positive integer)
    - `max_selections` is the maximum number of options that must be selected (positive integer)
    - `instructions` are the instructions for the question (string). If not provided, the default instructions are used. To view them, run `QuestionCheckBox.default_instructions`
    - `short_names_dict` maps question_options to short names (dictionary mapping strings to strings)

    For an example, run `QuestionCheckBox.example()`
    """

    question_type = "checkbox"
    question_options: list[str] = QuestionOptionsDescriptor()
    min_selections = IntegerDescriptor(none_allowed=True)
    max_selections = IntegerDescriptor(none_allowed=True)

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_options: list[str],
        min_selections: Optional[int] = None,
        max_selections: Optional[int] = None,
        short_names_dict: Optional[dict[str, str]] = None,
    ):
        self.question_name = question_name
        self.question_text = question_text
        self.min_selections = min_selections
        self.max_selections = max_selections
        self.question_options = question_options
        self.short_names_dict = short_names_dict or dict()

    ################
    # Answer methods
    ################
    def validate_answer(self, answer: Any) -> dict[str, Union[int, str]]:
        self.validate_answer_template_basic(answer)
        self.validate_answer_key_value(answer, "answer", list)
        self.validate_answer_checkbox(answer)
        return answer

    def translate_answer_code_to_answer(self, answer_codes, scenario: Scenario = None):
        """
        Translates the answer code to the actual answer.
        For example, for question options ["a", "b", "c"],the answer codes are 0, 1, and 2.
        The LLM will respond with [0,1] and this code will translate it to ["a","b"].
        """
        scenario = scenario or Scenario()
        translated_options = [
            Template(option).render(scenario) for option in self.question_options
        ]
        translated_codes = []
        for answer_code in answer_codes:
            translated_codes.append(translated_options[int(answer_code)])
        return translated_codes

    def simulate_answer(self, human_readable=True) -> dict[str, Union[int, str]]:
        """Simulates a valid answer for debugging purposes"""
        num_selections = random.randint(self.min_selections, self.max_selections)
        if human_readable:
            # Select a random number of options from self.question_options
            selected_options = random.sample(self.question_options, num_selections)
            answer = {
                "answer": selected_options,
                "comment": random_string(),
            }
        else:
            # Select a random number of indices from the range of self.question_options
            selected_indices = random.sample(
                range(len(self.question_options)), num_selections
            )
            answer = {
                "answer": selected_indices,
                "comment": random_string(),
            }
        return answer

    ################
    # Helpful methods
    ################
    @classmethod
    def example(cls) -> QuestionCheckBox:
        return cls(
            question_name="never_eat",
            question_text="Which of the following foods would you eat if you had to?",
            question_options=[
                "soggy meatpie",
                "rare snails",
                "mouldy bread",
                "panda milk custard",
                "McDonalds",
            ],
            min_selections=2,
            max_selections=5,
        )


def main():
    from edsl.questions.QuestionCheckBox import QuestionCheckBox

    q = QuestionCheckBox.example()
    q.question_text
    q.question_options
    q.question_name
    q.short_names_dict
    # validate an answer
    q.validate_answer({"answer": [1, 2], "comment": "I like custard"})
    # translate answer code
    q.translate_answer_code_to_answer([1, 2])
    # simulate answer
    q.simulate_answer()
    q.simulate_answer(human_readable=False)
    q.validate_answer(q.simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q



from edsl import Config
import sqlite3
from collections import UserList
from edsl.Base import Base
from rich.table import Table

config = Config()

path = config.get("EDSL_DATABASE_PATH")


class Cache(Base, UserList):
    def __init__(self, data=None, schema=None):
        self.db_path = config.get("EDSL_DATABASE_PATH")[len("sqlite:///") :]
        self.table_name = "responses"

        if data is None or schema is None:
            result = self.load_data()
            self.data = result["data"]
            self.schema = result["schema"]
        else:
            self.data = data
            self.schema = schema

    def _connect(self, db_path=None):
        return sqlite3.connect(db_path if db_path else self.db_path)

    def load_data(self):
        with self._connect() as conn:
            cur = conn.cursor()
            cur.execute(f"PRAGMA table_info({self.table_name})")
            schema = cur.fetchall()
            cur.execute(f"SELECT * FROM {self.table_name}")
            return {
                "data": [
                    dict(zip([col[1] for col in schema], row)) for row in cur.fetchall()
                ],
                "schema": schema,
            }

    def save_data_to_new_db(self, new_db_path):
        with self._connect(new_db_path) as new_conn:
            new_cur = new_conn.cursor()
            # Create table in new database
            columns = ", ".join([f"{col[1]} {col[2]}" for col in self.schema])
            new_cur.execute(f"CREATE TABLE IF NOT EXISTS {self.table_name} ({columns})")

            # Insert data into new table
            placeholders = ", ".join(["?"] * len(self.schema))
            new_cur.executemany(
                f"INSERT INTO {self.table_name} ({', '.join([col[1] for col in self.schema])}) VALUES ({placeholders})",
                [tuple(row[col[1]] for col in self.schema) for row in self.data],
            )
            new_conn.commit()

    def _table_row(self, row):
        table = Table()
        table.add_column("Key")
        table.add_column("Value")
        for key, value in row.items():
            table.add_row(str(key), str(value))
        return table

    def rich_print(self):
        table = Table(title="Cache")
        table.add_column("Entry", style="bold")
        for index, row in enumerate(self.data):
            table.add_row(self._table_row(row))
        return table

    @classmethod
    def example(cls):
        data0 = """{'id': 1, 'model': 'gpt-3.5-turbo'}"""
        schema = [
            (0, "id", "INTEGER", 1, None, 1),
            (1, "model", "VARCHAR(100)", 1, None, 0),
            (2, "parameters", "TEXT", 1, None, 0),
            (3, "system_prompt", "TEXT", 1, None, 0),
            (4, "prompt", "TEXT", 1, None, 0),
            (5, "output", "TEXT", 1, None, 0),
        ]
        return cls(data=[eval(data0)], schema=schema)

    def __repr__(self):
        return f"Cache(data={self.data}, schema={self.schema})"

    def code():
        pass

    def add_row(self, row_data):
        if set(row_data.keys()) != set([col[1] for col in self.schema]):
            raise ValueError("Row keys do not match table schema")
        self.data.append(row_data)

    def to_dict(self):
        return {"schema": self.schema, "data": self.data}

    @classmethod
    def from_dict(cls, data):
        raw_schema = data["schema"]
        schema = [tuple(col) for col in raw_schema]  # Convert to tuples
        cache = cls(data=data["data"], schema=schema)
        return cache


if __name__ == "__main__":
    cache = Cache()
    cache.print()
    # cache.load_data()



import textwrap
from abc import ABC
from typing import Any, List
import io

from rich.console import Console
from rich.table import Table

from jinja2 import Template, Environment, meta, TemplateSyntaxError

from edsl.exceptions.prompts import TemplateRenderError
from edsl.prompts.prompt_config import (
    C2A,
    names_to_component_types,
    ComponentTypes,
    NEGATIVE_INFINITY,
)
from edsl.prompts.registry import RegisterPromptsMeta

from edsl.Base import PersistenceMixin, RichPrintingMixin

MAX_NESTING = 100


class PromptBase(
    PersistenceMixin, RichPrintingMixin, ABC, metaclass=RegisterPromptsMeta
):
    component_type = ComponentTypes.GENERIC

    def __init__(self, text=None):
        if text is None:
            if hasattr(self, "default_instructions"):
                text = self.default_instructions
            else:
                text = ""
        self._text = text

    @property
    def text(self):
        return self._text

    def __add__(self, other_prompt):
        """
        >>> p = Prompt("Hello, {{person}}")
        >>> p2 = Prompt("How are you?")
        >>> p + p2
        Prompt(text='Hello, {{person}}How are you?')
        >>> p + "How are you?"
        Prompt(text='Hello, {{person}}How are you?')
        """
        if isinstance(other_prompt, str):
            return self.__class__(self.text + other_prompt)
        else:
            return self.__class__(text=self.text + other_prompt.text)

    def __str__(self):
        """
        >>> p = Prompt("Hello, {{person}}")
        >>> str(p)
        'Hello, {{person}}'
        """
        return self.text

    def __contains__(self, text_to_check):
        """
        >>> p = Prompt("Hello, {{person}}")
        >>> "person" in p
        True
        >>> "person2" in p
        False
        """
        return text_to_check in self.text

    def __repr__(self):
        """
        >>> p = Prompt("Hello, {{person}}")
        >>> p
        Prompt(text='Hello, {{person}}')
        """
        return f"Prompt(text='{self.text}')"

    def template_variables(
        self,
    ) -> list[str]:
        """
        >>> p = Prompt("Hello, {{person}}")
        >>> p.template_variables()
        ['person']
        """
        return self._template_variables(self.text)

    @staticmethod
    def _template_variables(template: str) -> list[str]:
        """ """
        env = Environment()
        ast = env.parse(template)
        return list(meta.find_undeclared_variables(ast))

    def undefined_template_variables(self, replcement_dict):
        return [var for var in self.template_variables() if var not in replcement_dict]

    def unused_traits(self, traits: dict):
        return [trait for trait in traits if trait not in self.template_variables()]

    @property
    def has_variables(self) -> bool:
        """
        >>> p = Prompt("Hello, {{person}}")
        >>> p.has_variables
        True
        >>> p = Prompt("Hello, person")
        >>> p.has_variables
        False
        """
        return len(self.template_variables()) > 0

    def render(self, primary_replacement: dict, **additional_replacements) -> str:
        """Renders the prompt with the replacements

        >>> p = Prompt("Hello, {{person}}")
        >>> p.render({"person": "John"})
        'Hello, John'
        >>> p.render({"person": "Mr. {{last_name}}", "last_name": "Horton"})
        'Hello, Mr. Horton'
        >>> p.render({"person": "Mr. {{last_name}}", "last_name": "Ho{{letter}}ton"}, max_nesting = 1)
        'Hello, Mr. Horton'
        """
        new_text = self._render(
            self.text, primary_replacement, **additional_replacements
        )
        return self.__class__(text=new_text)

    @staticmethod
    def _render(text, primary_replacement, **additional_replacements) -> "PromptBase":
        """
            Renders the template text with variables replaced from the provided named dictionaries.
            Allows for nested variable resolution up to a specified maximum nesting depth.

        >>> codebook = {"age": "Age"}
        >>> p = Prompt("You are an agent named {{ name }}. {{ codebook['age']}}: {{ age }}")
        >>> p.render({"name": "John", "age": 44}, codebook=codebook)
        'You are an agent named John. Age: 44'

        """
        try:
            previous_text = None
            for _ in range(MAX_NESTING):
                rendered_text = Template(text).render(
                    primary_replacement, **additional_replacements
                )
                if rendered_text == previous_text:
                    # No more changes, so return the rendered text
                    return rendered_text
                previous_text = text
                text = rendered_text

            # If the loop exits without returning, it indicates too much nesting
            raise TemplateRenderError(
                "Too much nesting - you created an infinite loop here, pal"
            )
        except TemplateSyntaxError as e:
            raise TemplateRenderError(f"Template syntax error: {e}")

    def to_dict(self):
        """
        >>> p = Prompt("Hello, {{person}}")
        >>> p.to_dict()
        {'text': 'Hello, {{person}}', 'class_name': 'Prompt'}
        """
        return {"text": self.text, "class_name": self.__class__.__name__}

    @classmethod
    def from_dict(cls, data):
        """
        >>> p = Prompt("Hello, {{person}}")
        >>> p2 = Prompt.from_dict(p.to_dict())
        >>> p2
        Prompt(text='Hello, {{person}}')
        """
        class_name = data["class_name"]
        cls = RegisterPromptsMeta._registry.get(class_name, Prompt)
        return cls(text=data["text"])

    def rich_print(self):
        """Displays an object as a table."""
        table = Table(title="Prompt")
        table.add_column("Attribute", style="bold")
        table.add_column("Value")

        to_display = self.__dict__.copy()
        for attr_name, attr_value in to_display.items():
            table.add_row(attr_name, repr(attr_value))
        table.add_row("Component type", str(self.component_type))
        table.add_row("Model", str(getattr(self, "model", "Not specified")))
        return table

    @classmethod
    def example(cls):
        return cls(cls.default_instructions)


class Prompt(PromptBase):
    component_type = ComponentTypes.GENERIC


from edsl.prompts.library.question_multiple_choice import *
from edsl.prompts.library.agent_instructions import *
from edsl.prompts.library.agent_persona import *

from edsl.prompts.library.question_budget import *
from edsl.prompts.library.question_checkbox import *
from edsl.prompts.library.question_freetext import *
from edsl.prompts.library.question_linear_scale import *
from edsl.prompts.library.question_numerical import *
from edsl.prompts.library.question_rank import *
from edsl.prompts.library.question_extract import *
from edsl.prompts.library.question_list import *


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import aiohttp
import json
from typing import Any
from edsl import CONFIG
from edsl.language_models.LanguageModel import LanguageModel
from edsl.enums import LanguageModelType, InferenceServiceType


class GeminiPro(LanguageModel):
    _inference_service_ = InferenceServiceType.GOOGLE.value
    _model_ = LanguageModelType.GEMINI_PRO.value
    _parameters_ = {
        "temperature": 0.5,
        "topP": 1,
        "topK": 1,
        "maxOutputTokens": 2048,
        "stopSequences": [],
        "use_cache": True,
    }
    api_token = CONFIG.get("GOOGLE_API_KEY")

    async def async_execute_model_call(
        self, user_prompt: str, system_prompt: str = ""
    ) -> dict[str, Any]:
        combined_prompt = user_prompt + system_prompt
        url = f"https://generativelanguage.googleapis.com/v1beta/models/gemini-pro:generateContent?key={self.api_token}"
        headers = {"Content-Type": "application/json"}
        data = {
            "contents": [{"parts": [{"text": combined_prompt}]}],
            "generationConfig": {
                "temperature": self.temperature,
                "topK": self.topK,
                "topP": self.topP,
                "maxOutputTokens": self.maxOutputTokens,
                "stopSequences": self.stopSequences,
            },
        }

        async with aiohttp.ClientSession() as session:
            async with session.post(
                url, headers=headers, data=json.dumps(data)
            ) as response:
                raw_response_text = await response.text()
                return json.loads(raw_response_text)

    def parse_response(self, raw_response: dict[str, Any]) -> str:
        data = raw_response
        return data["candidates"][0]["content"]["parts"][0]["text"]


if __name__ == "__main__":
    m = GeminiPro(use_cache=True)
    from edsl.questions import QuestionMultipleChoice

    q = QuestionMultipleChoice(
        question_text="Are pickled pigs feet a popular breakfast food in the US?",
        question_options=["Yes", "No", "Unsure"],
        question_name="bkfast_question",
    )
    results = q.by(m).run()
    from rich import print

    print(q)
    results.select("answer.*", "model.model").print()
    # results = m.execute_model_call("Why is Cape Cod a popular vacation spot?")
    # print(m.parse_response(results))



from __future__ import annotations

from abc import ABC, ABCMeta, abstractmethod
from edsl.jobs import Jobs
from edsl.results import Results


class RegisterJobsRunnerMeta(ABCMeta):
    "Metaclass to register output elements in a registry i.e., those that have a parent"
    _registry = {}  # Initialize the registry as a dictionary

    def __init__(cls, name, bases, dct):
        super(RegisterJobsRunnerMeta, cls).__init__(name, bases, dct)
        if name != "JobsRunner":
            RegisterJobsRunnerMeta._registry[name] = cls

    @classmethod
    def get_registered_classes(cls):
        return cls._registry

    @classmethod
    def lookup(cls):
        d = {}
        for classname, cls in cls._registry.items():
            if hasattr(cls, "runner_name"):
                d[cls.runner_name] = cls
            else:
                raise Exception(
                    f"Class {classname} does not have a runner_name attribute"
                )
        return d


class JobsRunner(ABC, metaclass=RegisterJobsRunnerMeta):
    """ABC for JobRunners, which take in a job, conduct interviews, and return their results."""

    def __init__(self, jobs: Jobs):
        self.jobs = jobs
        # create the interviews here so children can use them
        self.interviews = jobs.interviews()

    @abstractmethod
    def run(
        self,
        n: int = 1,
        debug: bool = False,
        verbose: bool = False,
        progress_bar: bool = True,
    ) -> Results:  # pragma: no cover
        """
        Runs the job: conducts Interviews and returns their results.
        - `n`: how many times to run each interview
        - `debug`: prints debug messages
        - `verbose`: prints messages
        - `progress_bar`: shows a progress bar
        """
        raise NotImplementedError



from __future__ import annotations
from collections.abc import Sequence
from itertools import product
from typing import Union
from edsl import CONFIG
from edsl.agents import Agent
from edsl.Base import Base
from edsl.data import Database, database
from edsl.language_models import LanguageModel, LanguageModelOpenAIThreeFiveTurbo
from edsl.results import Results
from edsl.scenarios import Scenario
from edsl.surveys import Survey
from edsl.jobs.base import JobsRunnersRegistry, JobsRunnerDescriptor
from edsl.jobs.Interview import Interview
from edsl.coop.old import JobRunnerAPI, ResultsAPI


class Jobs(Base):
    """
    The Jobs class is a collection of agents, scenarios and models and one survey.

    Methods:
    - `by()`: adds agents, scenarios or models to the job. Its a tricksy little method, be careful.
    - `interviews()`: creates a collection of interviews
    - `run()`: runs a collection of interviews

    """

    jobs_runner_name = JobsRunnerDescriptor()

    def __init__(
        self,
        survey: Survey,
        agents: list[Agent] = None,
        models: list[LanguageModel] = None,
        scenarios: list[Scenario] = None,
    ):
        self.survey = survey
        self.agents = agents or []
        self.models = models or []
        self.scenarios = scenarios or []

    def by(
        self,
        *args: Union[
            Agent,
            Scenario,
            LanguageModel,
            Sequence[Union[Agent, Scenario, LanguageModel]],
        ],
    ) -> Jobs:
        """
        Adds Agents, Scenarios and LanguageModels to a job. If no objects of this type exist in the Jobs instance, it stores the new objects as a list in the corresponding attribute. Otherwise, it combines the new objects with existing objects using the object's `__add__` method.

        Arguments:
        - objects or a sequence (list, tuple, ...) of objects of the same type

        Notes:
        - all objects must implement the 'get_value', 'set_value', and `__add__` methods
        - agents: traits of new agents are combined with traits of existing agents. New and existing agents should not have overlapping traits, and do not increase the # agents in the instance
        - scenarios: traits of new scenarios are combined with traits of old existing. New scenarios will overwrite overlapping traits, and do not increase the number of scenarios in the instance
        - models: new models overwrite old models.
        """
        # if the first argument is a sequence, grab it and ignore other arguments

        passed_objects = self._turn_args_to_list(args)

        current_objects, objects_key = self._get_current_objects_of_this_type(
            passed_objects[0]
        )

        if not current_objects:
            new_objects = passed_objects
        else:
            new_objects = self._merge_objects(passed_objects, current_objects)

        setattr(self, objects_key, new_objects)  # update the job
        return self

    @staticmethod
    def _turn_args_to_list(args):
        def did_user_pass_a_sequence(args):
            """
            >>> did_user_pass_a_sequence([1,2,3])
            True
            >>> did_user_pass_a_sequence(1)
            False
            """
            return len(args) == 1 and isinstance(args[0], Sequence)

        if did_user_pass_a_sequence(args):
            return list(args[0])
        else:
            return list(args)

    def _get_current_objects_of_this_type(self, object):
        class_to_key = {
            Agent: "agents",
            Scenario: "scenarios",
            LanguageModel: "models",
        }
        for class_type in class_to_key:
            if isinstance(object, class_type) or issubclass(
                object.__class__, class_type
            ):
                key = class_to_key[class_type]
                break
        else:
            raise ValueError(
                f"First argument must be an Agent, Scenario, or LanguageModel, not {object}"
            )
        current_objects = getattr(self, key, None)
        return current_objects, key

    @staticmethod
    def _merge_objects(passed_objects, current_objects):
        """
        Combines all the existing objects with the new objects
        For example, if the user passes in 3 agents,
        and there are 2 existing agents, this will create 6 new agents

        >>> Jobs(survey = [])._merge_objects([1,2,3], [4,5,6])
        [5, 6, 7, 6, 7, 8, 7, 8, 9]
        """
        new_objects = []
        for current_object in current_objects:
            for new_object in passed_objects:
                new_objects.append(current_object + new_object)
        return new_objects

    def interviews(self) -> list[Interview]:
        """
        Returns a list of Interviews, that will eventually be used by the JobRunner.
        - Returns one Interview for each combination of Agent, Scenario, and LanguageModel.
        - If any of Agents, Scenarios, or LanguageModels are missing, fills in with defaults. Note that this will change the corresponding class attributes.
        """
        self.agents = self.agents or [Agent()]
        self.models = self.models or [LanguageModelOpenAIThreeFiveTurbo(use_cache=True)]
        self.scenarios = self.scenarios or [Scenario()]
        interviews = []
        for agent, scenario, model in product(self.agents, self.scenarios, self.models):
            interview = Interview(
                survey=self.survey, agent=agent, scenario=scenario, model=model
            )
            interviews.append(interview)
        return interviews

    def run(
        self,
        n: int = 1,
        debug: bool = False,
        verbose: bool = False,
        progress_bar: bool = False,
        dry_run: bool = False,
        streaming: bool = False,
        db: Database = database,
    ) -> Union[Results, ResultsAPI, None]:
        """
        Runs the Job: conducts Interviews and returns their results.
        - `method`: "serial" or "threaded", defaults to "serial"
        - `n`: how many times to run each interview
        - `debug`: prints debug messages
        - `verbose`: prints messages
        - `progress_bar`: shows a progress bar
        """
        # self.job_runner_name = method
        if dry_run:
            self.job_runner_name = "dry_run"
        elif streaming:
            self.job_runner_name = "streaming"
        else:
            self.job_runner_name = "asyncio"

        if (emeritus_api_key := CONFIG.get("EMERITUS_API_KEY")) == "local":
            return self._run_local(
                n=n, verbose=verbose, debug=debug, progress_bar=progress_bar, db=db
            )
        else:
            results = self._run_remote(
                api_key=emeritus_api_key, job_dict=self.to_dict()
            )
        return results

    def _run_local(self, *args, db: Database = database, **kwargs):
        """Runs the job locally."""
        db._health_check_pre_run()
        JobRunner = JobsRunnersRegistry[self.job_runner_name](jobs=self)
        results = JobRunner.run(*args, **kwargs)
        db._health_check_post_run()
        return results

    def _run_remote(self, *args, **kwargs):
        """Runs the job remotely."""
        results = JobRunnerAPI(*args, **kwargs)
        return results

    #######################
    # Dunder methods
    #######################
    def __repr__(self) -> str:
        """Returns an eval-able string representation of the Jobs instance."""
        return f"Jobs(survey={repr(self.survey)}, agents={repr(self.agents)}, models={repr(self.models)}, scenarios={repr(self.scenarios)})"

    def __len__(self) -> int:
        """Returns the number of questions that will be asked while running this job."""
        number_of_questions = (
            len(self.agents or [1])
            * len(self.scenarios or [1])
            * len(self.models or [1])
            * len(self.survey)
        )
        return number_of_questions

    #######################
    # Serialization methods
    #######################
    def to_dict(self) -> dict:
        """Converts the Jobs instance to a dictionary."""
        return {
            "survey": self.survey.to_dict(),
            "agents": [agent.to_dict() for agent in self.agents],
            "models": [model.to_dict() for model in self.models],
            "scenarios": [scenario.to_dict() for scenario in self.scenarios],
        }

    @classmethod
    def from_dict(cls, data: dict) -> Jobs:
        """Creates a Jobs instance from a JSON string."""
        return cls(
            survey=Survey.from_dict(data["survey"]),
            agents=[Agent.from_dict(agent) for agent in data["agents"]],
            models=[LanguageModel.from_dict(model) for model in data["models"]],
            scenarios=[Scenario.from_dict(scenario) for scenario in data["scenarios"]],
        )

    #######################
    # Example methods
    #######################
    @classmethod
    def example(cls) -> Jobs:
        from edsl.questions import QuestionMultipleChoice

        from edsl import Agent

        # (status, question, period)
        agent_answers = {
            ("Joyful", "how_feeling", "morning"): "OK",
            ("Joyful", "how_feeling", "afternoon"): "Great",
            ("Joyful", "how_feeling_yesterday", "morning"): "Great",
            ("Joyful", "how_feeling_yesterday", "afternoon"): "Good",
            ("Sad", "how_feeling", "morning"): "Terrible",
            ("Sad", "how_feeling", "afternoon"): "OK",
            ("Sad", "how_feeling_yesterday", "morning"): "OK",
            ("Sad", "how_feeling_yesterday", "afternoon"): "Terrible",
        }

        def answer_question_directly(self, question, scenario):
            return agent_answers[
                (self.traits["status"], question.question_name, scenario["period"])
            ]

        sad_agent = Agent(traits = {"status": "Sad"})
        joy_agent = Agent(traits = {"status": "Joyful"})

        sad_agent.add_direct_question_answering_method(answer_question_directly)
        joy_agent.add_direct_question_answering_method(answer_question_directly)

        q1 = QuestionMultipleChoice(
            question_text="How are you this {{ period }}?",
            question_options=["Good", "Great", "OK", "Terrible"],
            question_name="how_feeling",
        )
        q2 = QuestionMultipleChoice(
            question_text="How were you feeling yesterday {{ period }}?",
            question_options=["Good", "Great", "OK", "Terrible"],
            question_name="how_feeling_yesterday",
        )
        base_survey = Survey(questions=[q1, q2])

        job = base_survey.by(
            Scenario({"period": "morning"}), Scenario({"period": "afternoon"})
        ).by(joy_agent, sad_agent)

        return job

    def rich_print(self):
        """Prints a rich representation of the Jobs instance."""
        from rich.table import Table

        table = Table(title="Jobs")
        table.add_column("Jobs")
        table.add_row(self.survey.rich_print())
        return table

    def code(self):
        raise NotImplementedError


def main():
    from edsl.jobs import Jobs

    job = Jobs.example()
    len(job) == 8
    results = job.run(debug=True)
    len(results) == 8
    results


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from enum import Enum


class EnumWithChecks(Enum):
    @classmethod
    def is_value_valid(cls, value):
        return any(value == item.value for item in cls)


class QuestionType(EnumWithChecks):
    MULTIPLE_CHOICE = "multiple_choice"
    YES_NO = "yes_no"
    FREE_TEXT = "free_text"
    RANK = "rank"
    BUDGET = "budget"
    CHECKBOX = "checkbox"
    EXTRACT = "extract"
    FUNCTIONAL = "functional"
    LIST = "list"
    NUMERICAL = "numerical"
    TOP_K = "top_k"
    LIKERT_FIVE = "likert_five"
    LINEAR_SCALE = "linear_scale"


# https://huggingface.co/meta-llama/Llama-2-70b-chat-hf


class LanguageModelType(EnumWithChecks):
    GPT_4 = "gpt-4-1106-preview"
    GPT_3_5_Turbo = "gpt-3.5-turbo"
    LLAMA_2_70B_CHAT_HF = "llama-2-70b-chat-hf"
    LLAMA_2_13B_CHAT_HF = "llama-2-13b-chat-hf"
    GEMINI_PRO = "gemini_pro"
    MIXTRAL_8x7B_INSTRUCT = "mixtral-8x7B-instruct-v0.1"
    TEST = "test"


class InferenceServiceType(EnumWithChecks):
    BEDROCK = "bedrock"
    DEEP_INFRA = "deep_infra"
    REPLICATE = "replicate"
    OPENAI = "openai"
    GOOGLE = "google"
    TEST = "test"



import platform
import subprocess
import tempfile

import copy
import base64
import functools
import inspect
import markdown2
import math
import matplotlib.pyplot as plt
import numpy as np
import os
import pandas as pd
import seaborn as sns
import string
import tempfile
import textwrap
import warnings
import webbrowser
from abc import ABC, abstractmethod, ABCMeta
from collections import Counter
from dataclasses import asdict
from io import BytesIO
from IPython.display import display, HTML
from scipy import stats
from scipy.stats import chisquare
from statsmodels.miscmodels.ordinal_model import OrderedModel
from statsmodels.tools.sm_exceptions import HessianInversionWarning, ConvergenceWarning
from typing import Callable
from wordcloud import WordCloud
from edsl.report.InputOutputDataTypes import (
    CategoricalData,
    NumericalData,
    PlotData,
    TallyData,
    CrossTabData,
    FreeTextData,
    ChiSquareData,
    RegressionData,
)
from edsl.utilities import is_notebook


def save_figure(filename):
    base, ext = os.path.splitext(filename)
    if ext.lower() == '.png':
        plt.savefig(filename, format='png')
    elif ext.lower() == '.jpg' or ext.lower() == '.jpeg':
        plt.savefig(filename, format='jpeg')
    elif ext.lower() == '.svg':
        plt.savefig(filename, format='svg')
    else:
        print("Unsupported file extension. Saving as PNG by default.")
        plt.savefig(base + '.png', format='png')

warnings.filterwarnings(
    "ignore",
    category=FutureWarning,
    module="seaborn._oldcore",
    message=".*is_categorical_dtype is deprecated.*",
)
warnings.filterwarnings("ignore", category=RuntimeWarning, module="statsmodels.*")
warnings.filterwarnings("ignore", category=RuntimeWarning, module="scipy.optimize.*")
warnings.filterwarnings("ignore", category=HessianInversionWarning)
warnings.filterwarnings("ignore", category=ConvergenceWarning)

def open_temp_file(file_path):
    system = platform.system()
    if system == 'Linux':
        subprocess.run(["xdg-open", file_path])
    elif system == 'Windows':
        os.startfile(file_path)
    elif system == 'Darwin':  # macOS
        subprocess.run(["open", file_path])
    else:
        print("Unsupported operating system")

def convert_svg_to_png_in_memory(svg_bytes):
    # Create a temporary SVG file
    with tempfile.NamedTemporaryFile(suffix=".svg", delete=False) as temp_svg:
        temp_svg.write(svg_bytes)
        open_temp_file(temp_svg.name)


class RegisterElementMeta(ABCMeta):
    "Metaclass to register output elements in a registry i.e., those that have a parent"
    _registry = {}  # Initialize the registry as a dictionary

    def __init__(cls, name, bases, dct):
        super(RegisterElementMeta, cls).__init__(name, bases, dct)
        if cls.LeftInputType is not None or cls.RightInputType is not None:
            # Register the class in the registry
            RegisterElementMeta._registry[name] = cls

    @classmethod
    def get_registered_classes(cls):
        return cls._registry


def camel_to_snake(name: str) -> str:
    """Converts a camel case string to snake case, e.g.,
    >>> camel_to_snake("HelloWorld")
    'hello_world'
    """
    snake_name = ""
    for index, char in enumerate(name):
        if char.isupper() and index != 0:
            snake_name += "_"
        snake_name += char.lower()

    return snake_name


class CustomFunctionWrapper:
    """A wrapper for a function that adds a name and docstring."""

    def __init__(self, func, name, doc):
        self._func = func
        self.name = name
        self.doc = doc

    def __call__(self, *args, **kwargs):
        return self._func(*args, **kwargs)

    def __repr__(self):
        return f"Method: `{self.name}`\nDescription: {self.doc or 'No description available'}"

    def _repr_html_(self):
        html = markdown2.markdown(
            f"**Method:** {self.name}\n\n**Description:** {self.doc or 'No description available'}"
        )
        # return markdown2.markdown(f"**Method:** {self.name}\n\n**Description:** {self.doc or 'No description available'}")
        # return f"<b>Method:</b> {self.name}<br><b>Description:</b> {self.doc or 'No description available'}"
        return html


def html_decorator(func: Callable) -> Callable:
    "A decorator that displays the output of a function as HTML."

    @functools.wraps(func)
    def wrapper(*args, **kwargs):        
        obj = func(*args, **kwargs)
        if is_notebook():  # if in a jupyter notebook
            html = obj.html()
            return display(HTML(html))
        else:
            return obj.view()  # otherwise open in a browser

    return wrapper


class Element(ABC, metaclass=RegisterElementMeta):
    """Base class for all elements.


    LeftInputType: The type of the left parent. Could be None.
    RightInputType: The type of the right parent. Could Be None.
    OutputDataType: The type of the output data.

    "Root" elements are those that do not have a parent, and are created from the results.

    """

    LeftInputType = None
    RightInputType = None
    OutputDataType = None

    def __init__(self, left_parent=None, right_parent=None, output_data=None, **kwargs):
        self.left_parent = left_parent
        self.right_parent = right_parent
        self.left_data = getattr(left_parent, "output_data", None)
        self.right_data = getattr(right_parent, "output_data", None)
        self.filename = None

        for key, value in kwargs.items():
            setattr(self, key, value)

        if (
            self.LeftInputType is not None
            and type(self.left_data) != self.LeftInputType
        ):
            raise TypeError(f"Left parent must be of type {self.LeftInputType}")

        if (
            self.RightInputType is not None
            and type(self.right_data) != self.RightInputType
        ):
            raise TypeError(f"Right parent must be of type {self.RightInputType}")

        if output_data is None:
            self.output_data = self.create_output(
                self.left_data, self.right_data, **kwargs
            )
        else:
            self.output_data = output_data

    @classmethod
    def unary(cls):
        print("Switch to using the cls.element_type method instead")
        return cls.RightInputType is None

    @property
    def data(self):
        print("Shift to using self.output_data")
        return self.output_data

    @classmethod
    @property
    def function_name(cls):
        return camel_to_snake(cls.__name__)

    @classmethod
    def element_type(cls):
        if cls.LeftInputType is None and cls.RightInputType is None:
            return "root"
        if cls.LeftInputType is not None and cls.RightInputType is None:
            return "unary"
        if cls.LeftInputType is not None and cls.RightInputType is not None:
            return "binary"

    @classmethod
    def code_generation(cls, results_name: str, left_column, right_column=None):
        if cls.element_type() == "unary":
            return f'{results_name}.{cls.function_name}("{left_column}")'
        elif cls.element_type() == "binary":
            return (
                f'{results_name}.{cls.function_name}("{left_column}", "{right_column}")'
            )
        elif cls.element_type() == "root":
            raise Exception("Should not be called on a root element")

    @abstractmethod
    def _primary_function(self):
        "The function that creates the output data, as a dictionary."
        raise NotImplementedError

    @abstractmethod
    def _html(self):
        "The function that creates the HTML representation of the output data"
        raise NotImplementedError

    def create_output(self, LeftInput, RightInput, **kwargs):
        if self.element_type() == "unary":
            output_data = self._primary_function(LeftInput, **kwargs)
        elif self.element_type() == "binary":
            output_data = self._primary_function(LeftInput, RightInput, **kwargs)
        elif self.element_type() == "root":
            raise Exception("Should not be called on a root element")
        else:
            raise Exception("Unknown element type")
        
        if output_data is None:
            self.filename = kwargs.get("filename", None)
            return None        

        return self.OutputDataType(**output_data)

    @classmethod
    def example(cls, **kwargs):
        class MockParent:
            def __init__(self, data):
                self.output_data = data

        left_parent = MockParent(cls.LeftInputType.example())
        right_parent = (
            None
            if cls.RightInputType is None
            else MockParent(cls.RightInputType.example())
        )

        return cls(left_parent, right_parent, **kwargs)

    def html(self):
        return self._html(**asdict(self.output_data))

    def view(self, **kwargs):
        if hasattr(self.output_data, "buffer"):     
            svg_bytes = self.output_data.buffer.getvalue()
            convert_svg_to_png_in_memory(svg_bytes)
        else:
            if self.filename:
                print(f"Output was written to file: {self.filename}")
            else:
                temporary_directory = tempfile.mkdtemp()
                with open(os.path.join(temporary_directory, "temp.html"), "w") as f:
                    f.write(self.html(**kwargs))
                webbrowser.open(os.path.join(temporary_directory, "temp.html"))

    @classmethod
    def parameters(cls):
        return inspect.signature(cls._primary_function).parameters

    @classmethod
    def create_external_function(cls, results) -> Callable:
        """Adds a function to the Results class that creates an output element.


        In ResultsOutputMixin, there is this function that iterates through the registered 
        classes and adds a function to the Results class for each one.

            def add_output_functions(self) -> None:
                output_classes = registery.get_registered_classes().values()
                self.analysis_options = []
                for output_class in output_classes:
                    new_function_name = output_class.function_name
                    new_function = output_class.create_external_function(self)
                    self.__dict__[new_function_name] = new_function

                    self.analysis_options.append({new_function_name: output_class.__doc__})

        """

        def create_parent(data_type, key, input_type):
            RootElement = create_root_element(input_type)
            parent = RootElement.from_results(results, key, input_type)
            return parent

        if cls.RightInputType is None:

            def func(column, **kwargs):
                left_parent = create_parent(
                    *results._parse_column(column), input_type=cls.LeftInputType
                )
                return cls(left_parent=left_parent, **kwargs)

        else:

            def func(left_column, right_column, **kwargs):
                left_parent = create_parent(
                    *results._parse_column(left_column), cls.LeftInputType
                )
                right_parent = create_parent(
                    *results._parse_column(right_column), cls.RightInputType
                )
                return cls(left_parent=left_parent, right_parent=right_parent, **kwargs)

        return CustomFunctionWrapper(
            html_decorator(func), doc=cls.help(), name=cls.function_name
        )

    @classmethod
    def help(cls):
        help_text = textwrap.dedent(
            f"""\
        {cls._primary_function.__doc__}
        """
        )
        # return self._primary_function.__doc__
        return help_text


def create_root_element(output_data_type):
    class Container(Element):
        LeftInputType = None
        RightInputType = None
        OutputDataType = output_data_type

        def _primary_function(self):
            raise Exception("Should not be called directly")

        @classmethod
        def from_results(cls, results, data_name, index=None):
            data_type, key = results._parse_column(data_name)
            output_data = results._fetch_element(data_type, key, cls.OutputDataType)
            return cls(
                name=data_name,
                left_parent=None,
                right_parent=None,
                output_data=output_data,
                index=index,
            )

        def _html(self):
            return self.output_data.html()

    return Container


class PlotMixin:
    OutputDataType = PlotData

    image_format = "svg"

    @staticmethod
    def plt_to_buf(plt, format=image_format):
        buf = BytesIO()
        plt.savefig(buf, format=format)
        buf.seek(0)
        plt.close()
        return buf

    def _html(
        self,
        buffer,
        title,
        format=image_format,
        option_codes=None,
        width_pct=100,
        **kwargs,
    ):
        image_base64 = base64.b64encode(buffer.getvalue()).decode("utf-8")
        html = []
        html.append(title)
        format_line = "png" if format == "png" else "svg+xml"
        html.append(
            f"""<img src="data:image/{format_line};base64,{image_base64}" style="width: {width_pct}%; height: auto;"  />"""
        )
        if option_codes is not None:
            left_option_codes = option_codes.get("left_option_codes", None)
            if left_option_codes is not None:
                html.append("<p>Codes:</p>")
                for key, value in left_option_codes.items():
                    html.append(f"<p><b>{value}</b>: {key}</p>")
            right_option_codes = option_codes.get("right_option_codes", None)
            if right_option_codes is not None:
                if right_option_codes is not None:
                    html.append("<p>2nd variable Codes:</p>")
                    for key, value in right_option_codes.items():
                        html.append(f"<p><b>{value}</b>: {key}</p>")
        return "\n".join(html)


def tally(responses, options):
    response_counts = dict(Counter(responses))
    for key in options:
        if key not in response_counts:
            response_counts[key] = 0
    return response_counts


def replace_with_alpha_codes(
    options: list[str], responses: list[str], prefix: str = ""
):
    code_gen = (chr(i) for i in range(65, 91))
    option_codes = {}
    for option in options:
        option_codes[option] = prefix + next(code_gen)
    new_options = [option_codes[option] for option in options]
    new_responses = [option_codes[response] for response in responses]
    return new_options, new_responses, option_codes


def header_version(options, index):
    cleaned_versions = [
        option.translate(str.maketrans("", "", string.punctuation)).lower()
        for option in options
    ]
    split_versions = [option.split(" ") for option in cleaned_versions]
    versions = [split_version[:index] for split_version in split_versions]
    return ["_".join(version) for version in versions]


def find_version(options, index):
    candidate = header_version(options, index)
    if len(set(candidate)) == len(candidate):
        return candidate
    else:
        return find_version(options, index + 1)


def get_option_codes_short_name(options):
    return dict(zip(options, find_version(options, 1)))


def to_strings(split_versions):
    return ["_".join(version) for version in split_versions]


def is_unique(split_versions):
    return len(set(to_strings(split_versions))) == len(to_strings(split_versions))


def get_option_codes(options: list[str]):
    """Creates a dictionary mapping options to codes."""
    cleaned_versions = [
        option.translate(str.maketrans("", "", string.punctuation)).lower()
        for option in options
    ]
    new_cleaned_versions = []
    shortshands = {"not sure": "not-sure", "need more": "need-more"}
    for option in cleaned_versions:
        candidate = copy.copy(option)
        for key, value in shortshands.items():
            candidate = candidate.replace(key, value)
            # print(f"Replacing {key} with {value}")
            # print(option)
            # print(option.replace(key, value))
        new_cleaned_versions.append(candidate)

    cleaned_versions = new_cleaned_versions
    split_versions = [option.split(" ") for option in cleaned_versions]
    # get rid of stop words, is possible
    stop_words = [
        "a",
        "an",
        "am",
        "the",
        "of",
        "and",
        "or",
        "to",
        "for",
        "in",
        "on",
        "im",
        "that",
        "are",
        "i",
    ]
    # removes stop words so long as doing so doesn't make responses non-unique
    for version in split_versions:
        for stop_word in stop_words:
            if stop_word in version:
                index = version.index(stop_word)
                del version[index]
                if not is_unique(split_versions):
                    version.insert(index, stop_word)
                    # break

    # starts at the end and pops off options if it keeps everything unique
    # if it goes through and doesn't pop anything for each option, it stops
    while True:
        num_pops = 0
        for version in split_versions:
            if len(version) > 1:  # we we get to one word, stop
                removed = version.pop()
                if is_unique(split_versions):  # no problem
                    pass
                else:  # oops, we cut into bone
                    version.append(removed)
                    num_pops += 1
            else:
                num_pops += 1
        if num_pops == len(
            split_versions
        ):  # stop the loop if we tried popping everyting w/ no luck
            break

    return dict(zip(options, ["_".join(version) for version in split_versions]))


def replace_with_codes(
    options: list[str], responses: list[str], short_names_dict=None, prefix=""
):
    if short_names_dict is not None:
        option_codes = short_names_dict
    else:
        option_codes = get_option_codes(options)

    new_options = [option_codes[option] for option in options]
    new_responses = [option_codes[response] for response in responses]
    return new_options, new_responses, option_codes


class BarChart(PlotMixin, Element):
    "Creates a bar chart plot for categorical data."
    LeftInputType = CategoricalData
    RightInputType = None

    def _primary_function(
        self,
        CategoricalDataObject,
        width=10,
        height=5,
        xlabel="Counts",
        ylabel="",
        footer_fontsize=8,
        title=None,
        use_code=None,
        width_pct=100,
        show_percentage=True,
        filename=None,
    ) -> dict:
        """
        Generates a bar chart from the provided categorical data object.

        ### Args:
            - CategoricalDataObject (CategoricalData): An object containing categorical data to be plotted.
            - `width (int, optional)`: Width of the plot. Defaults to 10.
            - height (int, optional): Height of the plot. Defaults to 5.
            - xlabel (str, optional): Label for the x-axis. Defaults to "Counts".
            - ylabel (str, optional): Label for the y-axis. Defaults to an empty string.
            - footer_fontsize (int, optional): Font size for the footer text. Defaults to 8.
            - title (str, optional): Title of the plot. If None, title is taken from CategoricalDataObject.text. Defaults to None.
            - use_code (bool, optional): Whether to use alphabetical codes for categorical options. Defaults to False.

        Note:
            If 'use_code' is set to True, each category in the plot is represented by an alphabetical code (A, B, C, ...),
            and a footer is added to the plot mapping these codes back to the original category names.
        """
        responses = CategoricalDataObject.responses
        options = CategoricalDataObject.options
        if title is None:
            title = CategoricalDataObject.text

        option_codes = None

        max_option_length = max([len(option) for option in options])
        if use_code is None:
            use_code = max_option_length > 10

        if use_code:
            if not (d := CategoricalDataObject.short_names_dict) == {}:
                options, responses, option_codes = replace_with_codes(
                    options, responses, short_names_dict=d
                )
            else:
                options, responses, option_codes = replace_with_codes(
                    options, responses
                )

        response_count = tally(responses, options)
        total_responses = sum(response_count.values())
        data = {key: response_count[key] for key in options}
        data_df = pd.DataFrame(list(data.items()), columns=["Keys", "Counts"])
        sns.set(style="whitegrid")
        plt.figure(figsize=(width, height))
        # sns.barplot(x="Counts", y="Keys", data=data_df, palette="Blues_d")
        # ax = sns.barplot(x="Counts", y="Keys", data=data_df, palette="Blues_d")
        ax = sns.barplot(
            x="Counts",
            y="Keys",
            data=data_df,
            palette="Blues_d",
            hue="Keys",
            legend=False,
        )

        # Adjust layout and add footer if necessary
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.title(f"{title}")

        if show_percentage:
            for p in ax.patches:
                percentage = f"{100 * p.get_width() / total_responses:.1f}%"  # Calculate percentage
                x = p.get_x() + p.get_width() + 0.5
                y = p.get_y() + p.get_height() / 2
                ax.text(x, y, percentage, ha="center", va="center")

        plt.tight_layout()

        if filename:
            save_figure(filename)
            return None

        return {
            "buffer": self.plt_to_buf(plt),
            "title": title,
            "option_codes": {
                "left_option_codes": option_codes,
                "right_option_codes": None,
            },
            "width_pct": width_pct,
        }


class HistogramPlot(PlotMixin, Element):
    LeftInputType = NumericalData
    RightInputType = None

    def _primary_function(
        self,
        NumericalDataObject,
        alpha=0.7,
        bins=30,
        xlabel="Value",
        ylabel="Frequency",
        color="blue",
        title=None,
        max_title_length=40,
        width_pct=100,
        filename=None,
    ):
        """
        Generates a histogram plot from a NumericalDataObject.

        This method plots a histogram based on the responses contained within the NumericalDataObject.
        It allows customization of the plot's appearance including the number of bins, transparency,
        color, and axis labels. Optionally, a custom title can be set, or it defaults to the 'text'
        attribute of the NumericalDataObject.

        Parameters:
            NumericalDataObject (NumericalData): An object containing numerical data and associated responses.
            alpha (float, optional): The transparency level of the histogram bars. Defaults to 0.7.
            bins (int, optional): The number of bins in the histogram. Defaults to 30.
            xlabel (str, optional): Label for the x-axis. Defaults to "Value".
            ylabel (str, optional): Label for the y-axis. Defaults to "Frequency".
            color (str, optional): Color of the histogram bars. Defaults to "blue".
            title (str, optional): Custom title for the histogram. If None, uses the 'text' attribute from NumericalDataObject.

        """
        responses = [
            float(x) if x is not None else None for x in NumericalDataObject.responses
        ]
        max_title_length = 40
        if title is None:
            if len(NumericalDataObject.text) > max_title_length:
                text = NumericalDataObject.text[:max_title_length] + "..."
            else:
                text = NumericalDataObject.text
        else:
            text = title
        plt.hist(responses, bins=bins, alpha=alpha, color=color)
        plt.title(f"{text}")
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.tight_layout()

        if filename is not None:
            save_figure(filename)
            return None

        return {
            "buffer": self.plt_to_buf(plt),
            "title": text,
            "option_codes": None,
            "width_pct": width_pct,
        }


class ScatterPlot(PlotMixin, Element):
    LeftInputType = NumericalData
    RightInputType = NumericalData

    def _primary_function(
        self,
        LeftNumericalDataObject,
        RightNumericalDataObject,
        alpha=0.5,
        title=None,
        regression_line=True,
        x_text=None,
        y_text=None,
        width_pct=100,
        filename = None, 
    ):
        """
        Generates a scatter plot using numerical data from two provided data objects.

        This method creates a scatter plot to visually represent the relationship between
        two sets of numerical data. It offers customization for the plot's transparency
        (alpha) and title.

        Args:
            LeftNumericalDataObject (NumericalData): The first numerical data object,
                                                        used for the x-axis data.
            RightNumericalDataObject (NumericalData): The second numerical data object,
                                                        used for the y-axis data.
            alpha (float, optional): The transparency level of the scatter plot points.
                                        A value between 0 (transparent) and 1 (opaque).
                                        Defaults to 0.5.
            title (str, optional): Title for the scatter plot. If None, a default title
                                    is generated using the text attributes of the
                                    NumericalData objects. Defaults to None.
        """
        x = LeftNumericalDataObject.responses
        y = RightNumericalDataObject.responses
        if x_text is None:
            x_text = LeftNumericalDataObject.text
        if y_text is None:
            y_text = RightNumericalDataObject.text

        if title is None:
            title = f"{x_text} vs {y_text}"

        plt.title("")
        plt.xlabel(x_text)
        plt.ylabel(y_text)
        plt.scatter(x, y, alpha=alpha)

        if regression_line:
            slope, intercept, r_value, p_value, std_err = stats.linregress(x, y)
            x_array = np.array(x)
            plt.plot(
                x, intercept + slope * x_array, color="red"
            )  # Plotting the regression line
            regression_info_text = (
                f"Slope: {slope:.3f}\n"
                f"Std Error in Slope: {std_err:.3f}\n"
                f"Intercept: {intercept:.2f}\n"
            )
            plt.text(
                0.05,
                0.95,
                regression_info_text,
                transform=plt.gca().transAxes,
                fontsize=9,
                verticalalignment="top",
                bbox=dict(boxstyle="round,pad=0.3", facecolor="white", alpha=0.5),
            )

        plt.tight_layout()

        if filename is not None:
            save_figure(filename)
            return None
        
        return {
            "buffer": self.plt_to_buf(plt),
            "title": "",
            "option_codes": None,
            "width_pct": width_pct,
        }


class WordCloudPlot(PlotMixin, Element):
    LeftInputType = FreeTextData
    RightInputType = None

    def _primary_function(
        self,
        FreeTextDataObject,
        width=800,
        height=400,
        background_color="white",
        width_pct=100, 
        filename = None
    ):
        """Creates a word cloud plot for free text data.

        Parameters
        ----------
        column: str
            Name of the column in the results to use.
        width : int
            Width of the plot in pixels.
        height : int
            Height of the plot in pixels.
        background_color : str
            Background color of the plot.
        """
        responses = " ".join(FreeTextDataObject.responses)
        text = FreeTextDataObject.text

        wordcloud = WordCloud(
            width=width, height=height, background_color=background_color
        ).generate(responses)
        plt.figure(figsize=(10, 5))
        plt.imshow(wordcloud, interpolation="bilinear")
        plt.axis("off")
        plt.title(f"{text}")

        if filename is not None:
            save_figure(filename)
            return None

        #
        #with open(filename, "w") as f:
        #    f.write(wordcloud)

        return {
            "buffer": self.plt_to_buf(plt),
            "title": "",
            "option_codes": None,
            "width_pct": width_pct,
        }


class Tally(Element):
    LeftInputType = CategoricalData
    RightInputType = None
    OutputDataType = TallyData

    def _primary_function(self, CategoricalDataObject, **kwargs):
        """Creates a tally of responses to a categorical question."""
        responses = CategoricalDataObject.responses
        text = CategoricalDataObject.text
        options = CategoricalDataObject.options

        response_count = dict(Counter(responses))
        # Add 0s for things that weren't selected even once
        for key in options:
            if key not in response_count:
                response_count[key] = 0

        options.reverse()
        return {
            "responses": {key: response_count[key] for key in options},
            "text": text,
        }

    def _html(self, responses, text, **kwargs):
        report_html = [
            "<div>",
            f"<p>{text}</p>" "<table>",
        ]
        for key, value in responses.items():
            report_html.append(f"<tr><td>{key}</td><td>{value}</td></tr>")
        report_html.append("</table>")
        report_html.append("</div>")
        return "\n".join(report_html)


def compute_cross_tab(left_responses, left_options, right_responses, right_options):
    left_response_count = dict(Counter(left_responses))
    right_response_count = dict(Counter(right_responses))
    # Add 0s for things that weren't selected even once
    for key in left_options:
        if key not in left_response_count:
            left_response_count[key] = 0
    for key in right_options:
        if key not in right_response_count:
            right_response_count[key] = 0

    left_options.reverse()
    right_options.reverse()

    cross_tab = {
        left_option: {right_option: 0 for right_option in right_options}
        for left_option in left_options
    }
    # Perform the cross-tabulation
    for left_response, right_response in zip(left_responses, right_responses):
        cross_tab[left_response][right_response] += 1
    return cross_tab


class CrossTab(Element):
    LeftInputType = CategoricalData
    RightInputType = CategoricalData
    OutputDataType = CrossTabData

    def _primary_function(
        self, LeftCategoricalDataObject, RightCategoricalDataObject, **kwargs
    ):
        """Creates a cross tabulation of two categorical variables.
        Parameters
        ----------
        left_column: str
            Name of the column in the results to use for the left side.
        right_column: str
            Name of the column in the results to use for the right side.
        """
        left_responses = LeftCategoricalDataObject.responses
        right_responses = RightCategoricalDataObject.responses
        left_text = LeftCategoricalDataObject.text
        right_text = RightCategoricalDataObject.text
        left_options = LeftCategoricalDataObject.options
        right_options = RightCategoricalDataObject.options

        cross_tab = compute_cross_tab(
            left_responses, left_options, right_responses, right_options
        )
        return {
            "cross_tab": cross_tab,
            "left_title": left_text,
            "right_title": right_text,
        }

    def _html(self, cross_tab, left_title, right_title, **kwargs):
        report_html = [
            "<div>",
            f"<p>Cross tabulation of: {left_title} and {right_title}</p>",
            "<table>",
        ]

        # Assuming all inner dictionaries have the same keys, use the keys from the first one
        first_key = next(iter(cross_tab))
        left_options = cross_tab[first_key].keys()
        headers = [""] + list(left_options)
        report_html.append(
            "<tr>" + "".join(f"<th>{header}</th>" for header in headers) + "</tr>"
        )

        # Fill in the rows of the table
        for right_option, counts in cross_tab.items():
            row = [f"<td>{right_option}</td>"]
            for left_option in left_options:
                row.append(f"<td>{counts[left_option]}</td>")
            report_html.append("<tr>" + "".join(row) + "</tr>")

        report_html.append("</table>")
        report_html.append("</div>")
        return "\n".join(report_html)


class FacetedBarChart(PlotMixin, Element):
    LeftInputType = CategoricalData
    RightInputType = CategoricalData

    def _primary_function(
        self,
        LeftCategoricalDataObject,
        RightCategoricalDataObject,
        num_cols=None,
        height=5,
        label_angle=45,
        title=None,
        use_code_left=None,
        use_code_right=None,
        sharey=True,
        width_pct=100,
        filename = None,
    ):
        """ "
            Generates a set of bar plots as a FacetGrid to compare two categorical data sets.

        This method creates a series of bar plots, one for each category in the RightCategoricalDataObject,
        to compare the frequencies of categories from LeftCategoricalDataObject. The plots are
        arranged in a grid layout, with an option to specify the number of columns and the height of each plot.
        Additionally, the angle of the x-axis labels and the title of the grid can be customized.

        Args:
            LeftCategoricalDataObject (CategoricalData): The first categorical data object,
                                                        used for the x-axis data in the bar plots.
            RightCategoricalDataObject (CategoricalData): The second categorical data object,
                                                        whose categories define the grid columns.
            num_cols (int, optional): The number of columns in the FacetGrid. If None, it's calculated
                                    based on the number of categories in RightCategoricalDataObject.
                                    Defaults to None.
            height (int, optional): The height of each subplot in the grid. Defaults to 5.
            label_angle (int, optional): The angle for rotating the x-axis labels for readability.
                                        Defaults to 45 degrees.
            title (str, optional): The overall title of the FacetGrid. If None, a default title is
                                generated based on the texts of the categorical data objects.
                                Defaults to None.
            use_code_left (bool, optional): Whether to use alphabetical codes for categorical options
                                            in the left data object.
            use_code_right (bool, optional): Whether to use alphabetical codes for categorical options
            sharey (bool, optional): Whether to share the y-axis across all plots. Defaults to True.

        Notes:
            - The bar plots are generated using seaborn's barplot function within a FacetGrid.
            - The layout of the grid is adjusted to accommodate the overall title and to prevent
            overlap of plot elements.
        """
        left_responses = LeftCategoricalDataObject.responses
        right_responses = RightCategoricalDataObject.responses
        left_text = LeftCategoricalDataObject.text
        right_text = RightCategoricalDataObject.text
        left_options = LeftCategoricalDataObject.options
        right_options = RightCategoricalDataObject.options

        if use_code_left is None:
            max_option_length_left = max([len(option) for option in left_options])
            use_code_left = max_option_length_left > 10
        if use_code_right is None:
            max_option_length_right = max([len(option) for option in right_options])
            use_code_right = max_option_length_right > 10

        if title is None:
            title = f'"{left_text}" \n by "{right_text}"'

        if len(left_text) > 40:
            left_text = left_text[:20] + "..."
        if len(right_text) > 40:
            right_text = right_text[:20] + "..."

        left_option_codes = None
        right_option_codes = None

        if use_code_left:
            left_options, left_responses, left_option_codes = replace_with_codes(
                left_options, left_responses, prefix="L-"
            )
        if use_code_right:
            right_options, right_responses, right_option_codes = replace_with_codes(
                right_options, right_responses, prefix="R-"
            )

        # Figures out how many columns to use in the FacetGrid if not specified
        if num_cols is None:
            if len(right_options) < 6:
                num_cols = len(right_options)
            else:
                num_cols = math.ceil(math.sqrt(len(right_options)))

        cross_tab = compute_cross_tab(
            right_responses, right_options, left_responses, left_options
        )

        d = {}
        if use_code_left:
            d = {v: k for k, v in left_option_codes.items()}
        left_option_name = d.get(left_text, left_text)
        if use_code_right:
            d = {v: k for k, v in right_option_codes.items()}
        right_option_name = d.get(right_text, right_text)

        df = pd.DataFrame(cross_tab)
        # Reset index to turn the index into a column
        df = df.reset_index()
        # Rename the columns to be more descriptive
        df.rename(columns={"index": left_option_name}, inplace=True)
        # Melt the DataFrame to long format
        df_long = df.melt(
            id_vars=left_option_name, var_name=right_option_name, value_name="Count"
        )
        sns.set(style="whitegrid")
        # Creating a FacetGrid
        g = sns.FacetGrid(
            df_long,
            col=right_option_name,
            col_wrap=num_cols,
            sharey=sharey,
            height=height,
        )
        # ax = sns.barplot(
        #     x="Counts",
        #     y="Keys",
        #     data=data_df,
        #     palette="Blues_d",
        #     hue="Keys",
        #     legend=False,
        # )
        # Adding bar plots to the FacetGrid
        g = g.map(
            sns.barplot,
            left_option_name,
            "Count",
            order=df_long[left_option_name].unique(),
            palette="viridis",
            hue=df_long[left_option_name].unique(),
            legend=False,
        )
        # Rotating x-axis labels for better readability
        for ax in g.axes.ravel():
            for label in ax.get_xticklabels():
                label.set_rotation(label_angle)

        g.fig.suptitle(f"{title}", fontsize=16)

        # Adjust the layout to make room for the title and prevent overlap
        g.fig.subplots_adjust(top=0.9)  # you can adjust the value as needed

        plt.tight_layout()

        if filename is not None:
            save_figure(filename)
            return None
        
        return {
            "buffer": self.plt_to_buf(plt),
            "title": "",
            "option_codes": {
                "left_option_codes": left_option_codes,
                "right_option_codes": right_option_codes,
            },
            "width_pct": width_pct,
        }


class ChiSquare(Element):
    LeftInputType = CategoricalData
    RightInputType = None
    OutputDataType = ChiSquareData

    def _primary_function(self, CategoricalDataObject, **kwargs):
        responses = CategoricalDataObject.responses
        text = CategoricalDataObject.text
        options = CategoricalDataObject.options

        response_count = dict(Counter(responses))
        # Add 0s for things that weren't selected even once
        for key in options:
            if key not in response_count:
                response_count[key] = 0

        observed_counts = list(response_count.values())
        chi_square, p_value = chisquare(observed_counts)
        return {"chi_square": chi_square, "p_value": p_value, "text": text}

    def _html(self, chi_square, p_value, text, digits=3, **kwargs):
        report_html = ["<div>", f"<p>Chi-square test for: {text}</p>" "<table>"]
        report_html.append(f"<p>Chi-square statistic: {round(chi_square, digits)}</p>")
        report_html.append(f"<p>p-value: {round(p_value, digits)}</p>")
        report_html.append("</div>")
        return "\n".join(report_html)


class OrderedLogit(Element):
    LeftInputType = CategoricalData
    RightInputType = CategoricalData
    OutputDataType = RegressionData

    def _primary_function(
        self, LeftSideCategoricalData, RightSideCategoricalData, **kwargs
    ):
        y = LeftSideCategoricalData.responses
        category_order = LeftSideCategoricalData.options
        X = RightSideCategoricalData.responses
        outcome_description = LeftSideCategoricalData.text
        if not (isinstance(y, list) and isinstance(X, list) and len(y) == len(X)):
            print(y)
            print(X)
            raise ValueError("y and X must be lists of the same length.")

        y_ordered = pd.Categorical(y, categories=category_order, ordered=True)

        # Create a DataFrame from the inputs
        data = pd.DataFrame({"Outcome": y_ordered, "Predictor": X})

        # Convert the categorical variable into dummy/indicator variables
        data = pd.get_dummies(data, columns=["Predictor"], drop_first=True)

        for col in data.columns.drop("Outcome"):
            data[col] = pd.to_numeric(data[col], errors="coerce")

        for col in data.select_dtypes(include=["bool"]).columns:
            data[col] = data[col].astype(int)
        try:
            model = OrderedModel(
                data["Outcome"], data.drop(columns=["Outcome"]), distr="logit"
            )  # Use 'logit' for logistic distribution
            result = model.fit()
            return {
                "model_outcome": result.summary().as_html(),
                "outcome_description": outcome_description,
            }
        except Exception as e:
            return {
                "model_outcome": f"Error: {e}",
                "outcome_description": outcome_description,
            }

    def _html(self, model_outcome: str, outcome_description: str):
        report_html = [
            "<h1>Ordered logit</h1>"
            "<div>",
            f"<p>Outcome: {outcome_description}</p>",
        ]
        report_html.append(model_outcome)
        report_html.append("</div>")
        return "\n".join(report_html)





from __future__ import annotations
import copy
import inspect
import types
import io
from typing import Any, Callable, Optional, Union, Dict

from jinja2 import Template

from rich.console import Console
from rich.table import Table

from edsl.Base import Base

from edsl.exceptions.agents import (
    AgentCombinationError,
    AgentDirectAnswerFunctionError,
    AgentDynamicTraitsFunctionError,
)

from edsl.agents.Invigilator import (
    InvigilatorDebug,
    InvigilatorHuman,
    InvigilatorFunctional,
    InvigilatorAI,
)

from edsl.language_models.registry import Model
from edsl.scenarios import Scenario
from edsl.enums import LanguageModelType

# from edsl.utilities import (
#     dict_to_html,
#     print_dict_as_html_table,
#     print_dict_with_rich,
# )

from edsl.agents.descriptors import (
    TraitsDescriptor,
    CodebookDescriptor,
    InstructionDescriptor,
    NameDescriptor
)

from edsl.utilities.decorators import sync_wrapper

from edsl.data_transfer_models import AgentResponseDict

from edsl.prompts.library.agent_persona import AgentPersona


class Agent(Base):
    """An agent that can answer questions.

    Parameters
    ----------
    traits : dict, optional - A dictionary of traits that the agent has. The keys need to be
    valid python variable names. The values can be any python object that has a valid __str__ method.
    codebook : dict, optional - A codebook mapping trait keys to trait descriptions.
    instruction : str, optional - Instructions for the agent.

    dynamic_traits_function : Callable, optional - A function that returns a dictionary of traits.

    """

    default_instruction = """You are answering questions as if you were a human. Do not break character."""

    traits = TraitsDescriptor()
    codebook = CodebookDescriptor()
    instruction = InstructionDescriptor()
    name = NameDescriptor()

    def __init__(
        self,
        *,
        traits: dict = None,
        name: str = None,
        codebook: dict = None,
        instruction: str = None,
        trait_presentation_template: str = None,
        dynamic_traits_function: Callable = None,
    ):
        self.name = name
        self._traits = traits or dict()
        self.codebook = codebook or dict()
        self.instruction = instruction or self.default_instruction
        self.dynamic_traits_function = dynamic_traits_function
        self._check_dynamic_traits_function()
        self.current_question = None

        if trait_presentation_template is not None:
            self.trait_presentation_template = trait_presentation_template
            self.agent_persona = AgentPersona(text=self.trait_presentation_template)

    def _check_dynamic_traits_function(self):
        if self.dynamic_traits_function:
            sig = inspect.signature(self.dynamic_traits_function)
            if "question" in sig.parameters:
                if len(sig.parameters) > 1:
                    raise AgentDynamicTraitsFunctionError(
                        f"The dynamic traits function {self.dynamic_traits_function} has too many parameters. It should only have one parameter: 'question'."
                    )
            else:
                if len(sig.parameters) > 0:
                    raise AgentDynamicTraitsFunctionError(
                        f"""The dynamic traits function {self.dynamic_traits_function} has too many parameters. It should have no parameters or 
                        just a single parameter: 'question'."""
                    )

    @property
    def traits(self) -> dict[str, str]:
        """A agent's traits, which is a dictionary.

        >> a = Agent(traits = {"age": 10, "hair": "brown", "height": 5.5})
        >> a.traits
        {'age': 10, 'hair': 'brown', 'height': 5.5}

        >> a = Agent()
        >> a.add_direct_question_answering_method(lambda question, scenario, self: {"age": 10, "hair": "brown", "height": 5.5})
        >> a.traits
        {'age': 10, 'hair': 'brown', 'height': 5.5}

        The agent could have a a dynamic traits function (dynamic_traits_function) that returns a dictionary of traits
        when called. This function can also take a question as an argument.  
        If so, the dynamic traits function is called and the result is returned.
        Otherwise, the traits are returned.
        """
        if self.dynamic_traits_function:
            sig = inspect.signature(self.dynamic_traits_function)
            if "question" in sig.parameters:
                return self.dynamic_traits_function(question=self.current_question)
            else:
                return self.dynamic_traits_function()
        else:
            return self._traits

    def add_direct_question_answering_method(self, method: Callable):
        """Adds a method to the agent that can answer a particular question type."""
        if hasattr(self, "answer_question_directly"):
            print("Warning: overwriting existing answer_question_directly method")

        signature = inspect.signature(method)
        for argument in ["question", "scenario", "self"]:
            if argument not in signature.parameters:
                raise AgentDirectAnswerFunctionError(
                    f"The method {method} does not have a '{argument}' parameter."
                )
        bound_method = types.MethodType(method, self)
        setattr(self, "answer_question_directly", bound_method)

    def create_invigilator(
        self,
        question: Question,
        scenario: Optional[Scenario] = None,
        model: Optional[LanguageModel] = None,
        debug: bool = False,
        memory_plan: Optional[MemoryPlan] = None,
        current_answers: Optional[dict] = None,
    ) -> 'Invigilator':
        """
        An invigator is an object that is responsible administering a question to an agent and
        recording the responses.
        """
        self.current_question = question
        model = model or Model(LanguageModelType.GPT_4.value, use_cache=True)
        scenario = scenario or Scenario()
        invigilator = self._create_invigilator(
            question, scenario, model, debug, memory_plan, current_answers
        )
        return invigilator

    async def async_answer_question(
        self,
        question: Question,
        scenario: Optional[Scenario] = None,
        model: Optional[LanguageModel] = None,
        debug: bool = False,
        memory_plan: Optional[MemoryPlan] = None,
        current_answers: Optional[dict] = None,
    ) -> AgentResponseDict:
        """
        This is a function where an agent returns an answer to a particular question.
        However, there are several different ways an agent can answer a question, so the
        actual functionality is delegated to an Invigilator object.
        """
        invigilator = self.create_invigilator(
            question=question,
            scenario=scenario,
            model=model,
            debug=debug,
            memory_plan=memory_plan,
            current_answers=current_answers,
        )
        response: AgentResponseDict = await invigilator.async_answer_question()
        return response

    answer_question = sync_wrapper(async_answer_question)

    def _create_invigilator(
        self,
        question: Question,
        scenario: Optional[Scenario] = None,
        model: Optional[LanguageModel] = None,
        debug: bool = False,
        memory_plan: Optional[MemoryPlan] = None,
        current_answers: Optional[dict] = None,
    ):
        model = model or Model(LanguageModelType.GPT_4.value, use_cache=True)
        scenario = scenario or Scenario()

        if debug:
            # use the question's simulate_answer method
            invigilator_class = InvigilatorDebug
        elif hasattr(question, "answer_question_directly"):
            # it is a functional question and the answer only depends on the agent's traits & the scenario
            invigilator_class = InvigilatorFunctional
        elif hasattr(self, "answer_question_directly"):
            # this of the case where the agent has a method that can answer the question directly
            # this occurrs when 'answer_question_directly' has been monkey-patched onto the agent
            # which happens when the agent is created from an existing survey
            invigilator_class = InvigilatorHuman
        else:
            # this means an LLM agent will be used. This is the standard case.
            invigilator_class = InvigilatorAI

        invigilator = invigilator_class(
            self, question, scenario, model, memory_plan, current_answers
        )
        return invigilator

    ################
    # Dunder Methods
    ################
    def __add__(self, other_agent: Agent = None) -> Agent:
        """
        Combines two agents by joining their traits.The agents must not have overlapping traits.
        >>> a1 = Agent(traits = {"age": 10})
        >>> a2 = Agent(traits = {"height": 5.5})
        >>> a1 + a2
        Agent(traits = {'age': 10, 'height': 5.5})
        >>> a1 + a1
        Traceback (most recent call last):
        ...
        edsl.exceptions.agents.AgentCombinationError: The agents have overlapping traits: {'age'}.
        """
        if other_agent is None:
            return self
        elif common_traits := set(self.traits.keys()) & set(other_agent.traits.keys()):
            raise AgentCombinationError(
                f"The agents have overlapping traits: {common_traits}."
            )
        else:
            new_agent = Agent(traits=copy.deepcopy(self.traits))
            new_agent.traits.update(other_agent.traits)
            return new_agent

    def __eq__(self, other: Agent) -> bool:
        """Checks if two agents are equal. Only checks the traits.
        >>> a1 = Agent(traits = {"age": 10})
        >>> a2 = Agent(traits = {"age": 10})
        >>> a1 == a2
        True
        >>> a3 = Agent(traits = {"age": 11})
        >>> a1 == a3
        False
        """
        return self.data == other.data

    def __repr__(self):
        class_name = self.__class__.__name__
        items = [
            f"{k} = '{v}'" if isinstance(v, str) else f"{k} = {v}"
            for k, v in self.data.items()
            if k != "question_type"
        ]
        return f"{class_name}({', '.join(items)})"

    ################
    # SERIALIZATION METHODS
    ################
    @property
    def data(self):
        raw_data = {
            k.replace("_", "", 1): v
            for k, v in self.__dict__.items()
            if k.startswith("_")
        }
        if hasattr(self, "set_instructions"):
            if not self.set_instructions:
                raw_data.pop("instruction")
        if self.codebook == {}:
            raw_data.pop("codebook")
        if self.name == None:
            raw_data.pop("name")
        return raw_data

    def to_dict(self) -> dict[str, Union[dict, bool]]:
        """Serializes to a dictionary."""
        return self.data

    @classmethod
    def from_dict(cls, agent_dict: dict[str, Union[dict, bool]]) -> Agent:
        """Deserializes from a dictionary."""
        return cls(**agent_dict)

    ################
    # DISPLAY Methods
    ################

    def _table(self) -> tuple[dict, list]:
        """Prepare generic table data."""
        table_data = []
        for attr_name, attr_value in self.__dict__.items():
            table_data.append({"Attribute": attr_name, "Value": repr(attr_value)})
        column_names = ["Attribute", "Value"]
        return table_data, column_names

    def rich_print(self):
        """Displays an object as a rich table."""
        table_data, column_names = self._table()
        table = Table(title=f"{self.__class__.__name__} Attributes")
        for column in column_names:
            table.add_column(column, style="bold")

        for row in table_data:
            row_data = [row[column] for column in column_names]
            table.add_row(*row_data)

        return table

    @classmethod
    def example(cls) -> Agent:
        """Returns an example agent.

        >>> Agent.example()
        Agent(traits = {'age': 22, 'hair': 'brown', 'height': 5.5})
        """
        return cls(traits={"age": 22, "hair": "brown", "height": 5.5})

    def code(self) -> str:
        """Returns the code for the agent."""
        return f"Agent(traits={self.traits})"


def main():
    """Consumes API credits"""
    from edsl.agents import Agent
    from edsl.questions import QuestionMultipleChoice

    # a simple agent
    agent = Agent(traits={"age": 10, "hair": "brown", "height": 5.5})
    agent.traits
    agent.print()
    # combining two agents
    agent = Agent(traits={"age": 10}) + Agent(traits={"height": 5.5})
    agent.traits
    # Agent -> Job using the to() method
    agent = Agent(traits={"allergies": "peanut"})
    question = QuestionMultipleChoice(
        question_text="Would you enjoy a PB&J?",
        question_options=["Yes", "No"],
        question_name="food_preference",
    )
    job = agent.to(question)
    # run the job
    results = job.run()
    # results


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from sqlalchemy import Column, String, Integer, Text, Index, Sequence
from sqlalchemy.orm import declarative_base

Base = declarative_base()


class LLMOutputDataDB(Base):
    __tablename__ = "responses"

    # Primary key
    id = Column(Integer, Sequence("response_id_seq"), primary_key=True)

    # Fields
    model = Column(String(100), nullable=False)
    parameters = Column(Text, nullable=False)
    system_prompt = Column(Text, nullable=False)
    prompt = Column(Text, nullable=False)
    output = Column(Text, nullable=False)

    # Index for faster queries
    idx_responses_fields = Index(
        "idx_responses_fields", "prompt", "system_prompt", "model", "parameters"
    )


class ResultDB(Base):
    __tablename__ = "result"

    id = Column(Integer, primary_key=True)
    job_uuid = Column(String, nullable=False)
    result_uuid = Column(String, nullable=False)
    agent = Column(Text, nullable=False)
    scenario = Column(Text, nullable=False)
    model = Column(Text, nullable=False)
    answer = Column(Text, nullable=False)

    idx_job_uuid = Index("idx_job_uuid", "job_uuid")



import json
import requests
from typing import Type
from edsl import CONFIG
from edsl.questions import Question

api_url = {
    "development": "http://127.0.0.1:8000",
    "production": "https://api.goemeritus.com",
}


class Coop:
    def __init__(self, api_key: str = None, run_mode: str = None) -> None:
        self.api_key = api_key or CONFIG.EMERITUS_API_KEY
        self.run_mode = run_mode or CONFIG.EDSL_RUN_MODE

    def __repr__(self):
        return f"Client(api_key='{self.api_key}', run_mode='{self.run_mode}')"

    @property
    def headers(self) -> dict:
        return {"Authorization": f"Bearer {self.api_key}"}

    @property
    def url(self) -> str:
        return api_url[self.run_mode]

    # QUESTIONS METHODS
    @property
    def questions(self) -> list[Type[Question]]:
        """Returns all questions the user has sent to the coop."""
        url = f"{self.url}/api/v0/questions/"
        response = requests.get(url, headers=self.headers).json()
        questions = [
            Question.from_dict(json.loads(q.get("json_string"))) for q in response
        ]
        return questions

    def get_question(self, question_id: int) -> Type[Question]:
        """Returns a question from the coop."""
        url = f"{self.url}/api/v0/questions/{question_id}"
        response = requests.get(url, headers=self.headers).json()
        if response.get("detail") == "Object not found":
            return None
        else:
            return Question.from_dict(json.loads(response.get("json_string")))

    def post_question(self, question: Type[Question]) -> dict:
        """Sends a question to the coop."""
        url = f"{self.url}/api/v0/questions"
        payload = {"json_string": json.dumps(question.to_dict())}
        response = requests.post(url, json=payload, headers=self.headers)
        return response.json()

    def delete_question(self, question_id: int) -> dict:
        """Deletes a question from the coop."""
        url = f"{self.url}/api/v0/questions/{question_id}"
        response = requests.delete(url, headers=self.headers)
        return response.json()

    # def send_job(self, job: Jobs) -> dict:
    #     """Sends a job to the server."""
    #     url = f"{self.url}/api/v0/jobs"
    #     payload = {"json_string": json.dumps(job.to_dict())}
    #     response = requests.post(url, json=payload, headers=self.headers)
    #     return response.json()

    # def run_job(self, job_id: str) -> dict:
    #     """Runs a job on the server."""
    #     url = f"{self.url}/api/v0/jobs/{job_id}/run"
    #     response = requests.post(url, headers=self.headers)
    #     return response.json()

    # def get_results(self, results_id: str) -> dict:
    #     """Returns the results of a job on the server."""
    #     url = f"{self.url}/api/v0/results/{results_id}"
    #     response = requests.get(url, headers=self.headers)
    #     return response.json()


if __name__ == "__main__":
    from edsl.coop import Coop
    from edsl.questions import QuestionMultipleChoice

    API_KEY = "mYdEgE3BFA4DG_ZH_UvFhUehgAEFgdZHEztCV3h_9bU"
    RUN_MODE = "development"
    coop = Coop(api_key=API_KEY, run_mode=RUN_MODE)

    # basics
    coop
    coop.headers
    coop.url

    # check jobs on server (should be an empty list)
    coop.questions

    # get a question that does not exist (should return None)
    coop.get_question(question_id=1)

    # now post a Question
    coop.post_question(QuestionMultipleChoice.example())

    # check all questions
    coop.questions

    # or get question by id
    coop.get_question(question_id=1)

    # delete the question
    coop.delete_question(question_id=1)

    # check all questions (should be an empty list)
    coop.questions



import re
import statsmodels.api as sm
from typing import List


def extract_variable_names(formula: str) -> List[str]:
    """
    Extracts the variable names from a formula
    TODO: improve the way we handle the case if there is a variable named 'C'
    """
    variable_names = re.findall(r"\b\w+\b", formula)
    return [variable_name for variable_name in variable_names if variable_name != "C"]


class RegressionMixin:
    def regression(
        self, formula: str
    ) -> sm.regression.linear_model.RegressionResultsWrapper:
        """Runs a linear regression"""
        variable_names = extract_variable_names(formula)
        df = self.select(*variable_names).to_pandas()
        # get rid of the headers
        column_mapping = {col: col.split(".")[-1] for col in df.columns}
        # Rename the columns using the dictionary
        df.rename(columns=column_mapping, inplace=True)
        model = sm.OLS.from_formula(formula, data=df)
        regression_result = model.fit()
        return regression_result



"""This tracker tracks API calls."""
import json
import time
from collections import namedtuple
from threading import Lock, Event
from edsl.trackers.Tracker import Tracker

UsageRates = namedtuple(
    "UsageRates",
    [
        "estimated_tokens_per_minute_k",
        "pct_of_tpm_limit",
        "estimated_requests_per_minute_k",
        "pct_of_rpm_limit",
    ],
)


class TrackerAPI(Tracker):
    """ """

    class APICallDetails(Tracker.TrackerEvent):
        def __init__(self, details):
            self.details = details

        def apply(self, tracker):
            with tracker.lock:
                tracker.usage.append(self.details)

    def __init__(
        self, lock, monitored_queue, tokens_per_minute_k=90, requests_per_minute_k=2
    ):
        self.lock = lock
        self.tokens_per_minute_k = tokens_per_minute_k
        self.requests_per_minute_k = requests_per_minute_k
        self.usage = []

        super().__init__(monitored_queue)

    def usage_rates(self, last_seconds=60):
        with self.lock:
            right_now = time.time()
            relevant_usage = [
                x
                for x in self.usage
                if (x["timestamp"] > right_now - last_seconds)
                and not x["cached_response"]
            ]
            if relevant_usage == []:
                return UsageRates(**{k: 0 for k in UsageRates._fields})
            else:
                elapsed_seconds = right_now - relevant_usage[0]["timestamp"]
                total_tokens = sum([x["total_tokens"] for x in relevant_usage])

                estimated_tokens_per_minute_k = round(
                    total_tokens / (elapsed_seconds / 60.0) / 1000, 0
                )
                pct_of_tpm_limit = round(
                    100 * estimated_tokens_per_minute_k / self.tokens_per_minute_k
                )

                estimated_requests_per_minute_k = round(
                    len(relevant_usage) / (elapsed_seconds / 60.0) / 1000, 2
                )
                pct_of_rpm_limit = round(
                    100 * estimated_requests_per_minute_k / self.requests_per_minute_k
                )

                usage_rates = UsageRates(
                    estimated_tokens_per_minute_k=estimated_tokens_per_minute_k,
                    pct_of_tpm_limit=pct_of_tpm_limit,
                    estimated_requests_per_minute_k=estimated_requests_per_minute_k,
                    pct_of_rpm_limit=pct_of_rpm_limit,
                )

                return usage_rates

    def tracked_values(self) -> dict:
        return {
            "Calls": len(self.usage),
        }

    def allowed_events(self):
        return set([TrackerAPI.APICallDetails])

    @classmethod
    def fromJSON(cls, filename):
        """This loads a tracker from a JSON file. Just for testing purposes"""
        with open(filename, "r") as f:
            json_dict = f.read()
        usage = json.loads(json_dict)
        # pretend responses are not cached for each testing, as the
        # usage_rate methods only compute w/ non-cached responses
        [x.update({"cached_response": False}) for x in usage]
        instance = cls(Lock(), None)
        instance.usage = usage
        return instance

    def toJSON(self, filename="sample_data.json"):
        with open(filename, "w") as f:
            json.dump(self.usage, f)

    def status(self):
        """This prints the status of the interview manager, while interviews are doing on."""
        estimated_length = 100
        completed = self.status_tracker.complete
        pct_completed = self.status_tracker.percentage_complete
        usage_rates = self.current_tpm(10)
        heartbeat = next(self.status_tracker.hearbeat)
        status_str = (
            f"{heartbeat} Completed: {completed} ({pct_completed} of total); "
            f"Est. TPM (k): {usage_rates.estimated_tokens_per_minute_k: ,} ({usage_rates.frac_of_tpm_limit}% of lim.); "
            f"Est. RPM (k): {usage_rates.estimated_requests_per_minute_k:,} ({usage_rates.frac_of_rpm_limit}% of lim.)); "
        )
        # Pad with spaces to reach the estimated length
        padded_status_str = status_str.ljust(estimated_length)
        print(f"\r{padded_status_str}", end="", flush=True)


if __name__ == "__main__":
    import textwrap
    from language_models import LanguageModelOpenAIThreeFiveTurbo

    m = LanguageModelOpenAIThreeFiveTurbo(use_cache=True)

    # the language model is attached to the queue
    print(m.api_queue)

    # the tracker expects a lock and an event
    lock = Lock()
    all_done = Event()

    # tracker attached to the queue
    tracker = TrackerAPI(lock=lock, monitored_queue=m.api_queue)

    # the queue is empty to start
    print(f"Queue size: {m.api_queue.qsize()}")

    # a call is made to the language model
    raw = m.get_raw_response(
        prompt="What is your favorite color?",
        system_prompt=textwrap.dedent(
            """\
                            You are pretending to be a human taking a survey.
                            Do not break character. 
                            """
        ),
    )

    # sleep for a second, as queue puts are non-blocking
    import time

    print("Sleeping for 1 second")
    time.sleep(1)
    # see that the queue updated
    print(f"Queue size: {m.api_queue.qsize()}")
    print(m.parse_response(raw))
    # indicate we are all done
    all_done.set()
    # have the tracker finish up
    tracker(all_done, interval=1)
    # how many calls did the API observe?
    print(tracker.tracked_values())
    # what is in the tracker queue?
    print(tracker.usage)
    # confirm that the API queue is empty
    print(f"Queue size: {m.api_queue.qsize()}")

    ## Next example - compute queue statistics
    all_done = Event()

    tracker_two = TrackerAPI(lock=lock, monitored_queue=m.api_queue)

    for i in range(10):
        raw = m.get_raw_response(
            prompt=f"What emotions are associated with {i}?",
            system_prompt=textwrap.dedent(
                """\
                                You are pretending to be a human taking a survey.
                                Do not break character. 
                                """
            ),
        )

    time.sleep(1)
    all_done.set()
    # have the tracker finish up
    tracker(all_done, interval=1)
    # print(tracker.usage)
    tracker.toJSON(filename="sample_data.json")

    # from .example_api_call_data import sample_data
    tracker = TrackerAPI.fromJSON("sample_data.json")
    print(tracker.usage)

    print(tracker.usage_rates(1000000))



from typing import List, Union
from collections import defaultdict, UserList

from edsl.exceptions import (
    SurveyRuleCannotEvaluateError,
    SurveyRuleCollectionHasNoRulesAtNodeError,
)
from edsl.utilities.interface import print_table_with_rich
from edsl.surveys.Rule import Rule
from edsl.surveys.base import EndOfSurvey
from edsl.surveys.DAG import DAG

from graphlib import TopologicalSorter

from collections import namedtuple

NextQuestion = namedtuple(
    "NextQuestion", "next_q, num_rules_found, expressions_evaluating_to_true, priority"
)

## We're going to need the survey object itself
## so we know how long the survey is, unless we move


class RuleCollection(UserList):
    "A collection of rules for a particular survey"

    def __init__(self, num_questions: int = None, rules: List[Rule] = None):
        super().__init__(rules or [])
        self.num_questions = num_questions

    def __repr__(self):
        """
        >>> rule_collection = RuleCollection.example()
        >>> _ = eval(repr(rule_collection))
        """
        return f"RuleCollection(rules={self.data}, num_questions={self.num_questions})"

    def to_dict(self):
        return {
            "rules": [rule.to_dict() for rule in self],
            "num_questions": self.num_questions,
        }

    @classmethod
    def from_dict(cls, rule_collection_dict):
        rules = [
            Rule.from_dict(rule_dict) for rule_dict in rule_collection_dict["rules"]
        ]
        num_questions = rule_collection_dict["num_questions"]
        new_rc = cls(rules=rules)
        new_rc.num_questions = num_questions
        return new_rc

    def add_rule(self, rule: Rule):
        """Adds a rule to a survey. If it's not, return human-readable complaints"""
        self.append(rule)

    def show_rules(self) -> None:
        keys = ["current_q", "expression", "next_q", "priority"]
        rule_list = []
        for rule in sorted(self, key=lambda r: r.current_q):
            rule_list.append({k: getattr(rule, k) for k in keys})

        print_table_with_rich(rule_list)

    def applicable_rules(self, q_now) -> list:
        """Which rules apply at the current node?

        >>> rule_collection = RuleCollection.example()
        >>> rule_collection.applicable_rules(1)
        [Rule(current_q=1, expression="q1 == 'yes'", next_q=3, priority=1, question_name_to_index={'q1': 1, 'q2': 2, 'q3': 3, 'q4': 4}), Rule(current_q=1, expression="q1 == 'no'", next_q=2, priority=1, question_name_to_index={'q1': 1, 'q2': 2, 'q3': 3, 'q4': 4})]

        More than one rule can apply. E.g., suppose we are at node 1.
        We could have three rules:
        1. "q1 == 'a' ==> 3
        2. "q1 == 'b' ==> 4
        3. "q1 == 'c' ==> 5
        """
        return [rule for rule in self if rule.current_q == q_now]

    def next_question(self, q_now, answers) -> NextQuestion:
        """Find the next question by index, given the rule collection"""
        # what rules apply at the current node?

        # tracking
        expressions_evaluating_to_true = 0
        next_q = None
        highest_priority = -2  # start with -2 to 'pick up' the default rule added
        num_rules_found = 0

        for rule in self.applicable_rules(q_now):
            num_rules_found += 1
            try:
                if rule.evaluate(answers):  # evaluates to True
                    expressions_evaluating_to_true += 1
                    if rule.priority > highest_priority:  # higher priority
                        # we have a new champ!
                        next_q, highest_priority = rule.next_q, rule.priority
            except SurveyRuleCannotEvaluateError:
                raise

        if num_rules_found == 0:
            raise SurveyRuleCollectionHasNoRulesAtNodeError(
                f"No rules found for question {q_now}"
            )

        return NextQuestion(
            next_q, num_rules_found, expressions_evaluating_to_true, highest_priority
        )

    @property
    def non_default_rules(self) -> List[Rule]:
        """Returns all rules that are not the default rule"
        >>> rule_collection = RuleCollection.example()
        >>> len(rule_collection.non_default_rules)
        2
        """
        return [rule for rule in self if rule.priority > -1]

    def keys_between(self, start_q, end_q, right_inclusive=True):
        """Returns a list of all question indices between start_q and end_q
        >>> rule_collection = RuleCollection(num_questions=5)
        >>> rule_collection.keys_between(1, 3)
        [2, 3]
        >>> rule_collection.keys_between(1, 4)
        [2, 3, 4]
        >>> rule_collection.keys_between(1, EndOfSurvey, right_inclusive=False)
        [2, 3]
        """

        # If it's the end of the survey, all questions between the start_q and the end of the survey
        # now depend on the start_q
        if end_q == EndOfSurvey:
            if self.num_questions is None:
                raise ValueError(
                    "Cannot determine DAG when EndOfSurvey and when num_questions is not known"
                )
            end_q = self.num_questions - 1

        question_range = list(range(start_q + 1, end_q + int(right_inclusive)))

        return question_range

    @property
    def dag(self) -> dict:
        """
        Finds the DAG of the survey, based on the skip logic.
        Keys are children questions; the list of values are nodes that must be answered first

        Rules are designated at the current question and then direct where
        control goes next. As such, the destination nodes are the keys
        and the current nodes are the values. Furthermore, all questions between
        the current and destination nodes are also included as keys, as they will depend
        on the answer to the focal node as well.

        ## If we have a rule that says "if q1 == 'yes', go to q3",
        ## Then q3 depends on q1, but so does q2
        ## So the DAG would be {3: [1], 2: [1]}

        >>> rule_collection = RuleCollection(num_questions=5)
        >>> qn2i = {'q1': 1, 'q2': 2, 'q3': 3, 'q4': 4}
        >>> rule_collection.add_rule(Rule(current_q=1, expression="q1 == 'yes'", next_q=3, priority=1,  question_name_to_index = qn2i))
        >>> rule_collection.add_rule(Rule(current_q=1, expression="q1 == 'no'", next_q=2, priority=1, question_name_to_index = qn2i))
        >>> rule_collection.dag
        {2: {1}, 3: {1}}
        """
        children_to_parents = defaultdict(set)
        # we are only interested in non-default rules. Default rules are those
        # that just go to the next question, so they don't add any dependencies
        for rule in self.non_default_rules:
            current_q, next_q = rule.current_q, rule.next_q
            for q in self.keys_between(current_q, next_q):
                children_to_parents[q].add(current_q)
        return DAG(dict(sorted(children_to_parents.items())))

    @classmethod
    def example(cls):
        qn2i = {"q1": 1, "q2": 2, "q3": 3, "q4": 4}
        return cls(
            num_questions=5,
            rules=[
                Rule(
                    current_q=1,
                    expression="q1 == 'yes'",
                    next_q=3,
                    priority=1,
                    question_name_to_index=qn2i,
                ),
                Rule(
                    current_q=1,
                    expression="q1 == 'no'",
                    next_q=2,
                    priority=1,
                    question_name_to_index=qn2i,
                ),
            ],
        )


if __name__ == "__main__":
    # pass
    import doctest

    doctest.testmod()



from collections import UserDict


class AgentResponseDict(UserDict):
    def __init__(self, *, question_name, answer, comment, prompts):
        super().__init__(
            {
                "answer": answer,
                "comment": comment,
                "question_name": question_name,
                "prompts": prompts,
            }
        )



import textwrap

from edsl.prompts.QuestionInstructionsBase import QuestionInstuctionsBase


class Rank(QuestionInstuctionsBase):
    question_type = "rank"
    model = "gpt-4-1106-preview"
    default_instructions = textwrap.dedent(
        """\
        You are being asked the following question: {{question_text}}
        The options are
        {% for option in question_options %}
        {{ loop.index0 }}: {{option}}
        {% endfor %}
        Return a valid JSON formatted like this, selecting the numbers of the options in order of preference,
        with the most preferred option first, and the least preferred option last:
        {"answer": [<put comma-separated list of answer codes here>], "comment": "<put explanation here>"}
        Exactly {{num_selections}} options must be selected.
        """
    )



import textwrap

from edsl.prompts.Prompt import PromptBase
from edsl.prompts.prompt_config import ComponentTypes

from edsl.enums import LanguageModelType


class AgentInstruction(PromptBase):
    model = LanguageModelType.GPT_3_5_Turbo.value
    component_type = ComponentTypes.AGENT_INSTRUCTIONS
    default_instructions = textwrap.dedent(
        """\
    You are playing the role of a human answering survey questions.
    Do not break character.
    """
    )


class AgentInstructionLlama(PromptBase):
    model = LanguageModelType.LLAMA_2_70B_CHAT_HF.value
    component_type = ComponentTypes.AGENT_INSTRUCTIONS
    default_instructions = textwrap.dedent(
        """\
    You are playing the role of a human answering questions.
    Do not break character.
    Only respond in JSON, with one answer formatted as specified.
    """
    )



from __future__ import annotations
from collections import UserDict
from typing import Any, Type

import io
from rich.console import Console
from rich.table import Table

from IPython.display import display

from edsl.agents import Agent
from edsl.language_models import LanguageModel
from edsl.scenarios import Scenario

from edsl.utilities import is_notebook

from edsl.Base import Base

def agent_namer_closure():
    """Returns a function that can be used to name an agent."""
    agent_dict = {}

    def agent_namer(agent):
        nonlocal agent_dict
        agent_count = len(agent_dict)
        if id(agent) in agent_dict:
            return agent_dict[id(agent)]
        else:
            agent_dict[id(agent)] = f"Agent_{agent_count}"
            return agent_dict[id(agent)]
 
    return agent_namer

agent_namer = agent_namer_closure()

class Result(Base, UserDict):
    """
    This class captures the result of one interview.
    - Its main data is an Agent, a Scenario, a Model, an Iteration, and an Answer.
    - These are stored both in the UserDict and as attributes.
    """

    def __init__(
        self,
        agent: Agent,
        scenario: Scenario,
        model: Type[LanguageModel],
        iteration: int,
        answer: str,
        prompt: dict[str, str] = None,
    ):
        # initialize the UserDict
        data = {
            "agent": agent,
            "scenario": scenario,
            "model": model,
            "iteration": iteration,
            "answer": answer,
            "prompt": prompt or {},
        }
        super().__init__(**data)
        # but also store the data as attributes
        self.agent = agent
        self.scenario = scenario
        self.model = model
        self.iteration = iteration
        self.answer = answer
        self.prompt = prompt or {}

    ###############
    # Used in Results
    ###############
    @property
    def sub_dicts(self) -> dict[str, dict]:
        """Returns a dictionary where keys are strings for each of the main class attributes/objects (except for iteration) and values are dictionaries for the attributes and values for each of these objects."""
        
        if self.agent.name is None:
            agent_name = agent_namer(self.agent)
        else:
            agent_name = self.agent.name

        return {
            "agent": self.agent.traits | {"agent_name": agent_name},
            "scenario": self.scenario,
            "model": self.model.parameters | {"model": self.model.model},
            "answer": self.answer,
            "prompt": self.prompt,
        }

    def code(self):
        raise NotImplementedError

    @property
    def combined_dict(self) -> dict[str, Any]:
        """Returns a dictionary that includes all sub_dicts, but also puts the key-value pairs in each sub_dict as a key_value pair in the combined dictionary."""
        combined = {}
        for key, sub_dict in self.sub_dicts.items():
            combined.update(sub_dict)
            combined.update({key: sub_dict})
        return combined

    def get_value(self, data_type: str, key: str) -> Any:
        """Returns the value for a given data type and key
        - data types can be "agent", "scenario", "model", or "answer"
        - keys are relevant attributes of the Objects the data types represent
        results.get_value("answer", "how_feeling") will return "Good" or "Bad" or whatnot
        """
        return self.sub_dicts[data_type][key]

    @property
    def key_to_data_type(self) -> dict[str, str]:
        """Returns a dictionary where keys are object attributes and values are the data type (object) that the attribute is associated with."""
        d = {}
        for data_type in ["agent", "scenario", "model", "answer", "prompt"]:
            for key in self.sub_dicts[data_type]:
                d[key] = data_type
        return d

    def rows(self, index):
        for data_type, subdict in self.sub_dicts.items():
            for key, value in subdict.items():
                yield (index, data_type, key, str(value))

    ###############
    # Useful
    ###############
    def copy(self) -> Result:
        """Returns a copy of the Result object."""
        return Result.from_dict(self.to_dict())

    def __eq__(self, other):
        return self.to_dict() == other.to_dict()

    ###############
    # Serialization
    ###############
    def to_dict(self) -> dict[str, Any]:
        """Returns a dictionary representation of the Result object."""
        return {
            k: v if not hasattr(v, "to_dict") else v.to_dict() for k, v in self.items()
        }

    @classmethod
    def from_dict(self, json_dict: dict) -> Result:
        """Returns a Result object from a dictionary representation."""
        result = Result(
            agent=Agent.from_dict(json_dict["agent"]),
            scenario=Scenario.from_dict(json_dict["scenario"]),
            model=LanguageModel.from_dict(json_dict["model"]),
            iteration=json_dict["iteration"],
            answer=json_dict["answer"],
            prompt=json_dict["prompt"],
        )
        return result

    def rich_print(self):
        """Displays an object as a table."""
        table = Table(title="Result")
        table.add_column("Attribute", style="bold")
        table.add_column("Value")

        to_display = self.__dict__.copy()
        data = to_display.pop("data", None)
        for attr_name, attr_value in to_display.items():
            table.add_row(attr_name, repr(attr_value))
        return table

    def __repr__(self):
        return f"Result(agent={repr(self.agent)}, scenario={repr(self.scenario)}, model={repr(self.model)}, iteration={self.iteration}, answer={repr(self.answer)}, prompt={repr(self.prompt)}"

    @classmethod
    def example(cls):
        from edsl.results import Results

        return Results.example()[0]


def main():
    from edsl.results.Result import Result
    import json

    print("Being imported")
    json_string = """
    {
        "agent": {
            "traits": {
                "status": "Unhappy"
            }
        },
        "scenario": {
            "period": "morning"
        },
        "model": {
            "model": "gpt-3.5-turbo",
            "parameters": {
                "temperature": 0.5,
                "max_tokens": 1000,
                "top_p": 1,
                "frequency_penalty": 0,
                "presence_penalty": 0,
                "use_cache": true
            }
        },
        "iteration": 0,
        "answer": {
            "how_feeling": "Bad"
        }, 
        "prompt": {"how_feeling_user_prompt": "How are you feeling today?", "how_feeling_system_prompt": "Answer the question"}
    }
    """

    result = Result.from_dict(json.loads(json_string))

    result.sub_dicts
    assert result.combined_dict["how_feeling"] == "Bad"

    result.combined_dict
    assert result.get_value("answer", "how_feeling") == "Bad"

    result.key_to_data_type
    print(result)

    assert result == result.copy()

    result.to_dict()


if __name__ == "__main__":
    print(Result.example())
